"""
Produce integration test data, which is tested by the `test_integration_features.py`
tests. One thing to note here is that all redshifts are reasonably high.

This is necessary, because low redshifts mean that neutral fractions are small,
and then numerical noise gets relatively more important, and can make the comparison
fail at the tens-of-percent level.
"""
import glob
import os
import sys

import h5py
from powerbox import get_power

from py21cmfast import AstroParams
from py21cmfast import CosmoParams
from py21cmfast import FlagOptions
from py21cmfast import UserParams
from py21cmfast import global_params
from py21cmfast import run_coeval
from py21cmfast import run_lightcone

SEED = 12345
DATA_PATH = os.path.join(os.path.dirname(__file__), "test_data")
DEFAULT_USER_PARAMS = {"HII_DIM": 50, "DIM": 150, "BOX_LEN": 100}
DEFAULT_ZPRIME_STEP_FACTOR = 1.04

OPTIONS = (
    [12, {}],
    [11, {"zprime_step_factor": 1.02}],
    [30, {"z_heat_max": 40}],
    [13, {"zprime_step_factor": 1.05, "z_heat_max": 25, "HMF": 0}],
    [16, {"interp_perturb_field": True}],
    [14, {"USE_MASS_DEPENDENT_ZETA": True}],
    [9, {"SUBCELL_RSD": True}],
    [10, {"INHOMO_RECO": True}],
    [16, {"HMF": 3, "USE_TS_FLUCT": True}],
    [20, {"z_heat_max": 45, "M_MIN_in_Mass": True, "HMF": 2}],
    [35, {"USE_FFTW_WISDOM": True}],
)


def get_defaults(kwargs, cls):
    return {k: kwargs.get(k, v) for k, v in cls._defaults_.items()}


def get_all_defaults(kwargs):
    flag_options = get_defaults(kwargs, FlagOptions)
    astro_params = get_defaults(kwargs, AstroParams)
    cosmo_params = get_defaults(kwargs, CosmoParams)
    user_params = get_defaults(kwargs, UserParams)
    return user_params, cosmo_params, astro_params, flag_options


def get_all_options(redshift, **kwargs):
    user_params, cosmo_params, astro_params, flag_options = get_all_defaults(kwargs)
    user_params.update(DEFAULT_USER_PARAMS)
    out = {
        "redshift": redshift,
        "user_params": user_params,
        "cosmo_params": cosmo_params,
        "astro_params": astro_params,
        "flag_options": flag_options,
        "regenerate": True,
        "write": False,
        "use_interp_perturb_field": kwargs.get("use_interp_perturb_field", False),
        "random_seed": SEED,
    }

    for key in kwargs:
        if key.upper() in (k.upper() for k in global_params.keys()):
            out[key] = kwargs[key]
    return out


def produce_coeval_power_spectra(redshift, **kwargs):
    options = get_all_options(redshift, **kwargs)

    coeval = run_coeval(**options)
    p, k = get_power(coeval.brightness_temp, boxlength=coeval.user_params.BOX_LEN,)

    return k, p, coeval


def produce_lc_power_spectra(redshift, **kwargs):
    options = get_all_options(redshift, **kwargs)

    lightcone = run_lightcone(max_redshift=options["redshift"] + 2, **options)

    p, k = get_power(
        lightcone.brightness_temp, boxlength=lightcone.lightcone_dimensions
    )

    return k, p, lightcone


def get_filename(redshift, **kwargs):
    # get sorted keys
    kwargs = {k: kwargs[k] for k in sorted(kwargs)}
    string = "_".join(f"{k}={v}" for k, v in kwargs.items())
    fname = f"power_spectra_z{redshift:.2f}_{string}.h5"

    return os.path.join(DATA_PATH, fname)


def produce_power_spectra_for_tests(redshift, **kwargs):
    k, p, coeval = produce_coeval_power_spectra(redshift, **kwargs)
    k_l, p_l, lc = produce_lc_power_spectra(redshift, **kwargs)

    fname = get_filename(redshift, **kwargs)

    # Need to manually remove it, otherwise h5py tries to add to it
    if os.path.exists(fname):
        os.remove(fname)

    with h5py.File(fname, "w") as fl:
        for k, v in kwargs.items():
            fl.attrs[k] = v

        fl.attrs["HII_DIM"] = coeval.user_params.HII_DIM
        fl.attrs["DIM"] = coeval.user_params.DIM
        fl.attrs["BOX_LEN"] = coeval.user_params.BOX_LEN

        fl["power_coeval"] = p
        fl["k_coeval"] = k

        fl["power_lc"] = p_l
        fl["k_lc"] = k_l

        fl["xHI"] = lc.global_xHI
        fl["Tb"] = lc.global_brightness_temp

    print(f"Produced {fname} with {kwargs}")


if __name__ == "__main__":
    global_params.ZPRIME_STEP_FACTOR = DEFAULT_ZPRIME_STEP_FACTOR

    for redshift, kwargs in OPTIONS:
        produce_power_spectra_for_tests(redshift, **kwargs)



"""
Unit tests for output structures
"""

import copy
import pickle

import pytest

from py21cmfast import InitialConditions  # An example of an output struct
from py21cmfast import IonizedBox
from py21cmfast import PerturbedField
from py21cmfast import TsBox
from py21cmfast import global_params


@pytest.fixture(scope="module")
def ic():
    return InitialConditions()


def test_arrays_inited(ic):
    assert not ic.arrays_initialized


def test_pointer_fields_ic(
    ic,
):  # TODO: this is probably good to implement for every output struct defined in code
    # Get list of fields before and after array initialisation
    d = copy.copy(list(ic.__dict__.keys()))
    ic._init_arrays()
    new_names = [name for name in ic.__dict__ if name not in d]

    assert new_names
    assert all(n in ic.pointer_fields for n in new_names)

    # cstruct shouldn't be initialised,
    assert not ic.arrays_initialized

    ic._init_cstruct()
    assert ic.arrays_initialized


def test_pointer_fields_pf():
    # Get list of fields before and after array initialisation

    with pytest.raises(KeyError):
        pf = PerturbedField()

    pf = PerturbedField(redshift=7.0)

    d = copy.copy(list(pf.__dict__.keys()))
    pf._init_arrays()
    new_names = [name for name in pf.__dict__ if name not in d]

    assert new_names
    assert all(n in pf.pointer_fields for n in new_names)


def test_pointer_fields_ib():
    # Get list of fields before and after array initialisation
    with pytest.raises(KeyError):
        pf = IonizedBox()

    pf = IonizedBox(redshift=7.0)

    d = copy.copy(list(pf.__dict__.keys()))
    pf._init_arrays()
    new_names = [name for name in pf.__dict__ if name not in d]

    assert new_names
    assert all(n in pf.pointer_fields for n in new_names)


def test_pointer_fields_st():
    # Get list of fields before and after array initialisation
    with pytest.raises(KeyError):
        pf = TsBox()

    pf = TsBox(redshift=7.0)

    d = copy.copy(list(pf.__dict__.keys()))
    pf._init_arrays()
    new_names = [name for name in pf.__dict__ if name not in d]

    assert new_names
    assert all(n in pf.pointer_fields for n in new_names)


def test_non_existence(ic, tmpdirec):
    assert not ic.exists(direc=tmpdirec.strpath)


def test_writeability(ic):
    with pytest.raises(IOError):
        ic.write()


def test_readability(tmpdirec):
    # we update this one, so don't use the global one
    ic_ = InitialConditions(init=True)

    # TODO: fake it being filled (need to do both of the following to fool it. Actually, we *shouldn't* be able to
    # TODO: fool it at all, but hey.
    ic_.filled = True
    ic_.random_seed  # accessing random_seed actually creates a random seed.

    ic_.write(direc=tmpdirec.strpath)

    ic2 = InitialConditions()

    assert (
        ic_._seedless_repr() == ic2._seedless_repr()
    )  # without seeds, they are obviously exactly the same.

    assert ic2.exists(direc=tmpdirec.strpath)

    ic2.read(direc=tmpdirec.strpath)

    assert repr(ic_) == repr(ic2)  # they should be exactly the same.
    assert str(ic_) == str(ic2)  # their str is the same.
    assert hash(ic_) == hash(ic2)
    assert ic_ == ic2
    assert ic_ is not ic2

    # make sure we can't read it twice
    with pytest.raises(IOError):
        ic2.read(direc=tmpdirec.strpath)


def test_different_seeds(ic):
    ic2 = InitialConditions(random_seed=1)

    assert ic is not ic2
    assert ic != ic2
    assert repr(ic) != repr(ic2)
    assert ic._seedless_repr() == ic2._seedless_repr()

    assert ic._md5 == ic2._md5

    # make sure we didn't inadvertantly set the random seed while doing any of this
    assert ic._random_seed is None


def test_pickleability():
    ic_ = InitialConditions(init=True)
    ic_.filled = True
    ic_.random_seed

    s = pickle.dumps(ic_)

    ic2 = pickle.loads(s)
    assert repr(ic_) == repr(ic2)


def test_fname():
    ic1 = InitialConditions(user_params={"HII_DIM": 1000})
    ic2 = InitialConditions(user_params={"HII_DIM": 1000})

    # we didn't give them seeds, so can't access the filename attribute (it is undefined until a seed is set)
    with pytest.raises(AttributeError):
        assert ic1.filename != ic2.filename  # random seeds are different

    # *but* should be able to get a skeleton filename:
    assert ic1._fname_skeleton == ic2._fname_skeleton

    ic1.random_seed  # sets the random seed
    ic2.random_seed

    assert ic1.filename != ic2.filename  # random seeds should now be different
    assert ic1._fname_skeleton == ic2._fname_skeleton


def test_match_seed(tmpdirec):
    # we update this one, so don't use the global one
    ic_ = InitialConditions(init=True, random_seed=12)

    # fake it being filled
    ic_.filled = True
    ic_.random_seed

    ic_.write(direc=tmpdirec.strpath)

    ic2 = InitialConditions(random_seed=1)
    with pytest.raises(
        IOError
    ):  # should not read in just anything if its random seed is set.
        ic2.read(direc=tmpdirec.strpath)


def test_bad_class_definition():
    class CustomInitialConditions(InitialConditions):
        _name = "InitialConditions"

        """
        A class containing all initial conditions boxes.
        """

        def _init_arrays(self):
            super()._init_arrays()

            # remove one of the arrays that needs to be defined.
            del self.hires_density

    with pytest.raises(AttributeError):
        CustomInitialConditions(init=True)


def test_pre_expose(ic):
    # haven't actually tried to read in or fill the array yet.
    with pytest.raises(Exception):
        ic._expose()


def test_bad_write():
    ic = InitialConditions()
    ic.filled = True

    # no random seed yet so shouldn't be able to write.
    with pytest.raises(ValueError):
        ic.write()


def test_global_params_keys():
    assert "HII_FILTER" in global_params.keys()



import pytest

from py21cmfast import global_params


@pytest.fixture(scope="session")
def tmpdirec(tmpdir_factory):
    """Pytest fixture instantiating a new session-scope "data" folder.

    Parameters
    ----------
    tmpdir_factory :
        Pytest fixture for creating temporary directories.
    """
    return tmpdir_factory.mktemp("data")


@pytest.fixture(autouse=True, scope="session")
def setup_and_teardown_package():
    # Set nice global defaults for testing purposes, to make runs faster
    # (can always be over-ridden per-test).
    original_zprime = global_params.ZPRIME_STEP_FACTOR
    global_params.ZPRIME_STEP_FACTOR = 1.2

    yield

    global_params.ZPRIME_STEP_FACTOR = original_zprime



"""A set of tools for reading/writing/querying the in-built cache."""
import glob
import logging
import os
from os import path

import h5py

from . import wrapper
from ._cfg import config
from .wrapper import global_params

logger = logging.getLogger("21cmFAST")


def readbox(*, direc=None, fname=None, hsh=None, kind=None, seed=None, load_data=True):
    """
    Read in a data set and return an appropriate object for it.

    Parameters
    ----------
    direc : str, optional
            The directory in which to search for the boxes. By default, this is the
            centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast/``.
    fname: str, optional
        The filename (without directory) of the data set. If given, this will be
        preferentially used, and must exist.
    hsh: str, optional
        The md5 hsh of the object desired to be read. Required if `fname` not given.
    kind: str, optional
        The kind of dataset, eg. "InitialConditions". Will be the name of a class
        defined in :mod:`~wrapper`. Required if `fname` not given.
    seed: str or int, optional
        The random seed of the data set to be read. If not given, and filename not
        given, then a box will be read if it matches the kind and hsh, with an
        arbitrary seed.
    load_data: bool, optional
        Whether to read in the data in the data set. Otherwise, only its defining
        parameters are read.

    Returns
    -------
    dataset :
        An output object, whose type depends on the kind of data set being read.

    Raises
    ------
    IOError :
        If no files exist of the given kind and hsh.
    ValueError :
        If either ``fname`` is not supplied, or both ``kind`` and ``hsh`` are not supplied.
    """
    direc = direc or path.expanduser(config["direc"])

    # We either need fname, or hsh and kind.
    if not fname and not (hsh and kind):
        raise ValueError("Either fname must be supplied, or kind and hsh")

    if fname:
        kind, hsh, seed = _parse_fname(fname)

    if not seed:
        fname = kind + "_" + hsh + "_r*.h5"
        files = glob.glob(path.join(direc, fname))
        if files:
            fname = files[0]
        else:
            raise IOError("No files exist with that kind and hsh.")
    else:
        fname = kind + "_" + hsh + "_r" + str(seed) + ".h5"

    # Now, open the file and read in the parameters
    with h5py.File(path.join(direc, fname), "r") as fl:
        # First get items out of attrs.
        top_level = {}
        for k, v in fl.attrs.items():
            top_level[k] = v

        # Now descend into each group of parameters
        params = {}
        for grp_nm, grp in fl.items():
            if grp_nm != kind:  # is a parameter
                params[grp_nm] = {}
                for k, v in grp.attrs.items():
                    params[grp_nm][k] = None if v == "none" else v

    # Need to map the parameters to input parameters.
    passed_parameters = {}
    for k, v in params.items():
        if "global_params" in k:
            for kk, vv in v.items():
                setattr(global_params, kk, vv)

        else:
            # The following line takes something like "cosmo_params", turns it into
            # "CosmoParams", and instantiates
            # that particular class with the dictionary parameters.
            passed_parameters[k] = getattr(wrapper, k.title().replace("_", ""))(**v)

    for k, v in top_level.items():
        passed_parameters[k] = v

    # Make an instance of the object.
    inst = getattr(wrapper, kind)(**passed_parameters)

    # Read in the actual data (this avoids duplication of reading data).
    if load_data:
        inst.read(direc=direc)

    return inst


def _parse_fname(fname):
    try:
        kind = fname.split("_")[0]
        hsh = fname.split("_")[1]
        seed = fname.split("_")[-1].split(".")[0][1:]
    except IndexError:
        raise ValueError("fname does not have correct format")

    if kind + "_" + hsh + "_r" + seed + ".h5" != fname:
        raise ValueError("fname does not have correct format")

    return kind, hsh, seed


def list_datasets(*, direc=None, kind=None, hsh=None, seed=None):
    """Yield all datasets which match a given set of filters.

    Can be used to determine parameters of all cached datasets, in conjunction with :func:`readbox`.

    Parameters
    ----------
    direc : str, optional
        The directory in which to search for the boxes. By default, this is the centrally-managed
        directory, given by the ``config.yml`` in ``.21cmfast``.
    kind: str, optional
        Filter by this kind (one of {"InitialConditions", "PerturbedField", "IonizedBox",
        "TsBox", "BrightnessTemp"}
    hsh: str, optional
        Filter by this hsh.
    seed: str, optional
        Filter by this seed.

    Yields
    ------
    fname: str
        The filename of the dataset (without directory).
    parts: tuple of strings
        The (kind, hsh, seed) of the data set.
    """
    direc = direc or path.expanduser(config["direc"])

    kind = kind or "*"
    hsh = hsh or "*"
    seed = seed or "*"

    fname = path.join(direc, str(kind) + "_" + str(hsh) + "_r" + str(seed) + ".h5")

    files = [path.basename(file) for file in glob.glob(fname)]

    for file in files:
        yield file, _parse_fname(file)


def query_cache(*, direc=None, kind=None, hsh=None, seed=None, show=True):
    """Get or print datasets in the cache.

    Walks through the cache, with given filters, and return all un-initialised dataset
    objects, optionally printing their representation to screen.
    Useful for querying which kinds of datasets are available within the cache, and
    choosing one to read and use.

    Parameters
    ----------
    direc : str, optional
        The directory in which to search for the boxes. By default, this is the
        centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast``.
    kind: str, optional
        Filter by this kind. Must be one of "InitialConditions", "PerturbedField",
        "IonizedBox", "TsBox" or "BrightnessTemp".
    hsh: str, optional
        Filter by this hsh.
    seed: str, optional
        Filter by this seed.
    show: bool, optional
        Whether to print out a repr of each object that exists.

    Yields
    ------
    obj:
       Output objects, un-initialized.
    """
    for file, parts in list_datasets(direc=direc, kind=kind, hsh=hsh, seed=seed):
        cls = readbox(direc=direc, fname=file, load_data=False)
        if show:
            print(file + ": " + str(cls))
        yield file, cls


def clear_cache(**kwargs):
    """Delete datasets in the cache.

    Walks through the cache, with given filters, and deletes all un-initialised dataset
    objects, optionally printing their representation to screen.

    Parameters
    ----------
    kwargs :
        All options passed through to :func:`query_cache`.
    """
    direc = kwargs.get("direc", path.expanduser(config["direc"]))
    number = 0
    for fname, cls in query_cache(show=False, **kwargs):
        if kwargs.get("show", True):
            logger.info("Removing {}".format(fname))
        os.remove(path.join(direc, fname))
        number += 1

    logger.info("Removed {} files from cache.".format(number))



from os import path

import pytest

import yaml

import py21cmfast as p21
from py21cmfast._cfg import Config


def test_config_context(tmpdirec):
    direc = tmpdirec.mkdir("config_test_dir")
    with p21.config.use(direc=direc, write=True):
        init = p21.initial_conditions(user_params={"HII_DIM": 30})

    assert path.exists(path.join(direc.strpath, init.filename))
    assert "config_test_dir" not in p21.config["direc"]


def test_config_write(tmpdirec):
    direc = tmpdirec.mkdir("config_write_dir")

    with p21.config.use(direc=direc):
        p21.config.write(path.join(direc, "config.yml"))

    with open(path.join(direc, "config.yml")) as fl:
        new_config = yaml.load(fl, Loader=yaml.FullLoader)

    # Test adding new kind of string alias
    new_config["boxdir"] = new_config["direc"]
    del new_config["direc"]

    with open(path.join(direc, "config.yml"), "w") as fl:
        yaml.dump(new_config, fl)

    with pytest.warns(UserWarning):
        new_config = Config.load(path.join(direc, "config.yml"))

    assert "boxdir" not in new_config
    assert "direc" in new_config

    with open(path.join(direc, "config.yml")) as fl:
        new_config = yaml.load(fl, Loader=yaml.FullLoader)

    assert "boxdir" not in new_config
    assert "direc" in new_config



"""The py21cmfast package."""
__version__ = "3.0.0.dev2"

# This just ensures that the default directory for boxes is created.
from os import mkdir as _mkdir
from os import path

from . import cache_tools
from ._cfg import config
from ._logging import configure_logging
from .cache_tools import query_cache
from .wrapper import AstroParams
from .wrapper import BrightnessTemp
from .wrapper import CosmoParams
from .wrapper import FlagOptions
from .wrapper import InitialConditions
from .wrapper import IonizedBox
from .wrapper import LightCone
from .wrapper import PerturbedField
from .wrapper import TsBox
from .wrapper import UserParams
from .wrapper import brightness_temperature
from .wrapper import compute_luminosity_function
from .wrapper import compute_tau
from .wrapper import global_params
from .wrapper import initial_conditions
from .wrapper import ionize_box
from .wrapper import perturb_field
from .wrapper import run_coeval
from .wrapper import run_lightcone
from .wrapper import spin_temperature

configure_logging()

try:
    _mkdir(path.expanduser(config["direc"]))
except FileExistsError:
    pass



"""
Output class objects.

The classes provided by this module exist to simplify access to large datasets created within C.
Fundamentally, ownership of the data belongs to these classes, and the C functions merely accesses
this and fills it. The various boxes and quantities associated with each output are available as
instance attributes. Along with the output data, each output object contains the various input
parameter objects necessary to define it.

.. warning:: These should not be instantiated or filled by the user, but always handled
             as output objects from the various functions contained here. Only the data
             within the objects should be accessed.
"""
import numpy as np
from cached_property import cached_property

from ._utils import OutputStruct as _BaseOutputStruct
from .c_21cmfast import ffi
from .inputs import AstroParams
from .inputs import CosmoParams
from .inputs import FlagOptions
from .inputs import UserParams
from .inputs import global_params


class _OutputStruct(_BaseOutputStruct):
    _global_params = global_params

    def __init__(self, *, user_params=None, cosmo_params=None, **kwargs):
        if cosmo_params is None:
            cosmo_params = CosmoParams()
        if user_params is None:
            user_params = UserParams()

        super().__init__(user_params=user_params, cosmo_params=cosmo_params, **kwargs)

    _ffi = ffi


class _OutputStructZ(_OutputStruct):
    _inputs = _OutputStruct._inputs + ["redshift"]


class InitialConditions(_OutputStruct):
    """A class containing all initial conditions boxes."""

    # The filter params indicates parameters to overlook when deciding if a cached box
    # matches current parameters.
    # It is useful for ignoring certain global parameters which may not apply to this
    # step or its dependents.
    _filter_params = _OutputStruct._filter_params + [
        "ALPHA_UVB",  # ionization
        "EVOLVE_DENSITY_LINEARLY",  # perturb
        "SMOOTH_EVOLVED_DENSITY_FIELD",  # perturb
        "R_smooth_density",  # perturb
        "HII_ROUND_ERR",  # ionization
        "FIND_BUBBLE_ALGORITHM",  # ib
        "N_POISSON",  # ib
        "T_USE_VELOCITIES",  # bt
        "MAX_DVDR",  # bt
        "DELTA_R_HII_FACTOR",  # ib
        "HII_FILTER",  # ib
        "INITIAL_REDSHIFT",  # pf
        "HEAT_FILTER",  # st
        "CLUMPING_FACTOR",  # st
        "Z_HEAT_MAX",  # st
        "R_XLy_MAX",  # st
        "NUM_FILTER_STEPS_FOR_Ts",  # ts
        "ZPRIME_STEP_FACTOR",  # ts
        "TK_at_Z_HEAT_MAX",  # ts
        "XION_at_Z_HEAT_MAX",  # ts
        "Pop",  # ib
        "Pop2_ion",  # ib
        "Pop3_ion",  # ib
        "NU_X_BAND_MAX",  # st
        "NU_X_MAX",  # ib
    ]

    def _init_arrays(self):
        self.lowres_density = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )
        self.lowres_vx = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.lowres_vy = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.lowres_vz = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.lowres_vx_2LPT = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )
        self.lowres_vy_2LPT = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )
        self.lowres_vz_2LPT = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )
        self.hires_density = np.zeros(
            self.user_params.tot_fft_num_pixels, dtype=np.float32
        )
        self.hires_vcb = np.zeros(self.user_params.tot_fft_num_pixels, dtype=np.float32)
        self.lowres_vcb = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )

        shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )
        hires_shape = (self.user_params.DIM, self.user_params.DIM, self.user_params.DIM)

        self.lowres_density.shape = shape
        self.lowres_vx.shape = shape
        self.lowres_vy.shape = shape
        self.lowres_vz.shape = shape
        self.lowres_vx_2LPT.shape = shape
        self.lowres_vy_2LPT.shape = shape
        self.lowres_vz_2LPT.shape = shape
        self.hires_density.shape = hires_shape

        self.lowres_vcb.shape = shape
        self.hires_vcb.shape = hires_shape


class PerturbedField(_OutputStructZ):
    """A class containing all perturbed field boxes."""

    _filter_params = _OutputStruct._filter_params + [
        "ALPHA_UVB",  # ionization
        "HII_ROUND_ERR",  # ionization
        "FIND_BUBBLE_ALGORITHM",  # ib
        "N_POISSON",  # ib
        "T_USE_VELOCITIES",  # bt
        "MAX_DVDR",  # bt
        "DELTA_R_HII_FACTOR",  # ib
        "HII_FILTER",  # ib
        "HEAT_FILTER",  # st
        "CLUMPING_FACTOR",  # st
        "Z_HEAT_MAX",  # st
        "R_XLy_MAX",  # st
        "NUM_FILTER_STEPS_FOR_Ts",  # ts
        "ZPRIME_STEP_FACTOR",  # ts
        "TK_at_Z_HEAT_MAX",  # ts
        "XION_at_Z_HEAT_MAX",  # ts
        "Pop",  # ib
        "Pop2_ion",  # ib
        "Pop3_ion",  # ib
        "NU_X_BAND_MAX",  # st
        "NU_X_MAX",  # ib
    ]

    def _init_arrays(self):
        self.density = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.velocity = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)

        self.density.shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )
        self.velocity.shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )


class _AllParamsBox(_OutputStructZ):
    _inputs = _OutputStructZ._inputs + ["flag_options", "astro_params"]

    _filter_params = _OutputStruct._filter_params + [
        "T_USE_VELOCITIES",  # bt
        "MAX_DVDR",  # bt
    ]

    def __init__(self, astro_params=None, flag_options=None, first_box=False, **kwargs):
        if flag_options is None:
            flag_options = FlagOptions()

        if astro_params is None:
            astro_params = AstroParams(INHOMO_RECO=flag_options.INHOMO_RECO)

        self.first_box = first_box

        super().__init__(astro_params=astro_params, flag_options=flag_options, **kwargs)


class IonizedBox(_AllParamsBox):
    """A class containing all ionized boxes."""

    def _init_arrays(self):
        # ionized_box is always initialised to be neutral for excursion set algorithm.
        # Hence np.ones instead of np.zeros
        self.xH_box = np.ones(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.Gamma12_box = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )
        self.z_re_box = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.dNrec_box = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)

        shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )
        self.xH_box.shape = shape
        self.Gamma12_box.shape = shape
        self.z_re_box.shape = shape
        self.dNrec_box.shape = shape

    @cached_property
    def global_xH(self):
        """Global (mean) neutral fraction."""
        if not self.filled:
            raise AttributeError(
                "global_xH is not defined until the ionization calculation has been performed"
            )
        else:
            return np.mean(self.xH_box)


class TsBox(_AllParamsBox):
    """A class containing all spin temperature boxes."""

    def _init_arrays(self):
        self.Ts_box = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.x_e_box = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.Tk_box = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)

        self.Ts_box.shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )
        self.x_e_box.shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )
        self.Tk_box.shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )

    @cached_property
    def global_Ts(self):
        """Global (mean) spin temperature."""
        if not self.filled:
            raise AttributeError(
                "global_Ts is not defined until the ionization calculation has been performed"
            )
        else:
            return np.mean(self.Ts_box)

    @cached_property
    def global_Tk(self):
        """Global (mean) Tk."""
        if not self.filled:
            raise AttributeError(
                "global_Tk is not defined until the ionization calculation has been performed"
            )
        else:
            return np.mean(self.Tk_box)

    @cached_property
    def global_x_e(self):
        """Global (mean) x_e."""
        if not self.filled:
            raise AttributeError(
                "global_x_e is not defined until the ionization calculation has been performed"
            )
        else:
            return np.mean(self.x_e_box)


class BrightnessTemp(_AllParamsBox):
    """A class containing the brightness temperature box."""

    _filter_params = _OutputStructZ._filter_params

    def _init_arrays(self):
        self.brightness_temp = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )

        self.brightness_temp.shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )

    @cached_property
    def global_Tb(self):
        """Global (mean) brightness temperature."""
        if not self.filled:
            raise AttributeError(
                "global_Tb is not defined until the ionization calculation has been performed"
            )
        else:
            return np.mean(self.brightness_temp)



"""
Various tests of the initial_conditions() function and InitialConditions class.
"""

import pytest

import numpy as np

from py21cmfast import wrapper


@pytest.fixture(scope="module")  # call this fixture once for all tests in this module
def basic_init_box():
    return wrapper.initial_conditions(
        regenerate=True, write=False, user_params=wrapper.UserParams(HII_DIM=35)
    )


def test_box_shape(basic_init_box):
    """Test basic properties of the InitialConditions struct"""
    shape = (35, 35, 35)
    assert basic_init_box.lowres_density.shape == shape
    assert basic_init_box.lowres_vx.shape == shape
    assert basic_init_box.lowres_vy.shape == shape
    assert basic_init_box.lowres_vz.shape == shape
    assert basic_init_box.lowres_vx_2LPT.shape == shape
    assert basic_init_box.lowres_vy_2LPT.shape == shape
    assert basic_init_box.lowres_vz_2LPT.shape == shape
    assert basic_init_box.hires_density.shape == tuple(3 * s for s in shape)

    assert basic_init_box.cosmo_params == wrapper.CosmoParams()


def test_modified_cosmo():
    """Test using a modified cosmology"""
    cosmo = wrapper.CosmoParams(sigma_8=0.9)
    ic = wrapper.initial_conditions(
        cosmo_params=cosmo,
        regenerate=True,
        write=False,
        user_params={"HII_DIM": 35, "BOX_LEN": 70},
    )

    assert ic.cosmo_params == cosmo
    assert ic.cosmo_params.SIGMA_8 == cosmo.SIGMA_8


def test_transfer_function(basic_init_box):
    """Test using a modified transfer function"""
    ic = wrapper.initial_conditions(
        regenerate=True,
        write=False,
        random_seed=basic_init_box.random_seed,
        user_params=wrapper.UserParams(HII_DIM=35, BOX_LEN=70, POWER_SPECTRUM=5),
    )

    rmsnew = np.sqrt(np.mean(ic.hires_density ** 2))
    rmsdelta = np.sqrt(np.mean((ic.hires_density - basic_init_box.hires_density) ** 2))
    assert rmsdelta < rmsnew
    assert rmsnew > 0.0
    assert not np.allclose(ic.hires_density, basic_init_box.hires_density)


def test_relvels():
    """Test for relative velocity initial conditions"""
    ic = wrapper.initial_conditions(
        regenerate=True,
        write=False,
        random_seed=1,
        user_params=wrapper.UserParams(
            HII_DIM=100,
            DIM=300,
            BOX_LEN=300,
            POWER_SPECTRUM=5,
            USE_RELATIVE_VELOCITIES=True,
        ),
    )

    vcbrms = np.sqrt(np.mean(ic.hires_vcb ** 2))
    vcbavg = np.mean(ic.hires_vcb)

    vcbrms_lowres = np.sqrt(np.mean(ic.lowres_vcb ** 2))
    vcbavg_lowres = np.mean(ic.lowres_vcb)

    assert vcbrms > 25.0
    assert vcbrms < 35.0  # it should be about 30 km/s, so we check it is around it

    # the average should be 0.92*vrms, since it follows a maxwell boltzmann
    assert vcbavg < 0.95 * vcbrms
    assert vcbavg > 0.90 * vcbrms

    # we also test the lowres box. Increase its error margin because its low res.
    assert vcbrms_lowres > 20.0
    assert vcbrms_lowres < 40.0
    assert vcbavg_lowres < 0.97 * vcbrms_lowres
    assert vcbavg_lowres > 0.88 * vcbrms_lowres



"""Simple script to simulate an MCMC-like routine to check for memory leaks."""
import os
import tracemalloc

import numpy as np
import psutil

from py21cmfast import initial_conditions
from py21cmfast import perturb_field
from py21cmfast import run_coeval

tracemalloc.start()
snapshot = tracemalloc.take_snapshot()
PROCESS = psutil.Process(os.getpid())
oldmem = 0


def trace_print():
    """Print a trace of memory leaks."""
    global snapshot
    global oldmem

    snapshot2 = tracemalloc.take_snapshot()
    snapshot2 = snapshot2.filter_traces(
        (
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<unknown>"),
            tracemalloc.Filter(False, tracemalloc.__file__),
        )
    )

    if snapshot is not None:
        thismem = PROCESS.memory_info().rss / 1024 ** 2
        diff = thismem - oldmem
        print(
            "===================== Begin Trace (TOTAL MEM={:1.4e} MB... [{:+1.4e} MB]):".format(
                thismem, diff
            )
        )
        top_stats = snapshot2.compare_to(snapshot, "lineno", cumulative=True)
        for stat in top_stats[:4]:
            print(stat)
        print("End Trace ===========================================")
        oldmem = thismem

    snapshot = snapshot2


NITER = 50

init = initial_conditions(
    user_params={"HII_DIM": 50, "BOX_LEN": 125.0}, regenerate=True
)
perturb = (
    perturb_field(redshift=7, init_boxes=init),
    perturb_field(redshift=8, init_boxes=init),
    perturb_field(redshift=9, init_boxes=init),
)

astro_params = {"HII_EFF_FACTOR": np.random.normal(30, 0.1)}

for i in range(NITER):
    trace_print()

    coeval = run_coeval(
        redshift=[7, 8, 9],
        astro_params=astro_params,
        init_box=init,
        perturb=perturb,
        regenerate=True,
        random_seed=init.random_seed,
        write=False,
    )



"""
Unit tests for input structures
"""

import pickle

import pytest

from py21cmfast import AstroParams  # An example of a struct with defaults
from py21cmfast import CosmoParams
from py21cmfast import FlagOptions
from py21cmfast import UserParams
from py21cmfast import global_params


@pytest.fixture(scope="module")
def c():
    return CosmoParams(SIGMA_8=0.8)


def test_diff(c):
    "Ensure that the python dict has all fields"
    d = CosmoParams(SIGMA_8=0.9)

    assert c is not d
    assert c != d
    assert hash(c) != hash(d)


def test_constructed_the_same(c):
    c2 = CosmoParams(SIGMA_8=0.8)

    assert c is not c2
    assert c == c2
    assert hash(c) == hash(c2)


def test_bad_construction(c):
    with pytest.raises(TypeError):
        CosmoParams(c, c)

    with pytest.raises(TypeError):
        CosmoParams(UserParams())

    with pytest.raises(TypeError):
        CosmoParams(1)


def test_warning_bad_params(caplog):
    CosmoParams(bad_param=1)
    assert (
        "The following parameters to CosmoParams are not supported: ['bad_param']"
        in caplog.text
    )


def test_constructed_from_itself(c):
    c3 = CosmoParams(c)

    assert c == c3
    assert c is not c3


def test_dynamic_variables():
    u = UserParams()
    assert u.DIM == 3 * u.HII_DIM

    u.update(DIM=200)

    assert u.DIM == 200


def test_repr(c):
    assert "SIGMA_8:0.8" in repr(c)


def test_pickle(c):
    # Make sure we can pickle/unpickle it.
    c4 = pickle.dumps(c)
    c4 = pickle.loads(c4)
    assert c == c4

    # Make sure the c data gets loaded fine.
    assert c4._cstruct.SIGMA_8 == c._cstruct.SIGMA_8


def test_self(c):
    c5 = CosmoParams(c.self)
    assert c5 == c
    assert c5.pystruct == c.pystruct
    assert c5.defining_dict == c.defining_dict
    assert (
        c5.defining_dict != c5.pystruct
    )  # not the same because the former doesn't include dynamic parameters.
    assert c5.self == c.self


def test_update():
    c = CosmoParams()
    c_pystruct = c.pystruct

    c.update(
        SIGMA_8=0.9
    )  # update c parameters. since pystruct as dynamically created, it is a new object each call.
    assert c_pystruct != c.pystruct


def test_c_structures(c):
    # See if the C structures are behaving correctly
    c2 = CosmoParams(SIGMA_8=0.8)

    assert c() != c2()
    assert (
        c() is c()
    )  # Re-calling should not re-make the object (object should have persistence)


def test_c_struct_update():
    c = CosmoParams()
    _c = c()
    c.update(SIGMA_8=0.8)
    assert _c != c()


def test_update_inhomo_reco(caplog):
    ap = AstroParams(R_BUBBLE_MAX=25)

    ap.update(INHOMO_RECO=True)

    msg = (
        "You are setting R_BUBBLE_MAX != 50 when INHOMO_RECO=True. "
        + "This is non-standard (but allowed), and usually occurs upon manual update of INHOMO_RECO"
    )

    ap.R_BUBBLE_MAX

    assert msg in caplog.text


def test_mmin():
    fo = FlagOptions(USE_MASS_DEPENDENT_ZETA=True)
    assert fo.M_MIN_in_Mass


def test_globals():
    orig = global_params.Z_HEAT_MAX

    with global_params.use(Z_HEAT_MAX=10.0):
        assert global_params.Z_HEAT_MAX == 10.0
        assert global_params._cobj.Z_HEAT_MAX == 10.0

    assert global_params.Z_HEAT_MAX == orig



"""
Testing plots is kind of hard, but we just check that it runs through without crashing.
"""

import pytest

from py21cmfast import initial_conditions
from py21cmfast import plotting
from py21cmfast import run_lightcone


def test_coeval_sliceplot():
    ic = initial_conditions(user_params={"HII_DIM": 35, "DIM": 70})

    fig, ax = plotting.coeval_sliceplot(ic)

    assert ax.xaxis.get_label().get_text() == "x-axis [Mpc]"
    assert ax.yaxis.get_label().get_text() == "y-axis [Mpc]"

    with pytest.raises(ValueError):  # bad slice axis
        plotting.coeval_sliceplot(ic, slice_axis=-2)

    with pytest.raises(IndexError):  # tring to plot slice that doesn't exist
        plotting.coeval_sliceplot(ic, slice_index=50)

    fig2, ax2 = plotting.coeval_sliceplot(ic, fig=fig, ax=ax)

    assert fig2 is fig
    assert ax2 is ax

    fig2, ax2 = plotting.coeval_sliceplot(ic, fig=fig)

    assert fig2 is fig
    assert ax2 is ax

    fig2, ax2 = plotting.coeval_sliceplot(ic, ax=ax)

    assert fig2 is fig
    assert ax2 is ax

    fig, ax = plotting.coeval_sliceplot(
        ic, kind="hires_density", slice_index=50, slice_axis=1
    )
    assert ax.xaxis.get_label().get_text() == "x-axis [Mpc]"
    assert ax.yaxis.get_label().get_text() == "z-axis [Mpc]"


def test_lightcone_sliceplot():
    lc = run_lightcone(
        redshift=25, max_redshift=30, user_params={"HII_DIM": 35, "DIM": 70}
    )

    fig, ax = plotting.lightcone_sliceplot(lc)

    assert ax.yaxis.get_label().get_text() == "y-axis [Mpc]"
    assert ax.xaxis.get_label().get_text() == "Redshift Axis [Mpc]"



"""Utilities that help with wrapping various C structures."""
import glob
import logging
import warnings
from hashlib import md5
from os import path

import h5py
import numpy as np
from cffi import FFI

from ._cfg import config

_ffi = FFI()

logger = logging.getLogger("21cmFAST")


class StructWrapper:
    """
    A base-class python wrapper for C structures (not instances of them).

    Provides simple methods for creating new instances and accessing field names and values.

    To implement wrappers of specific structures, make a subclass with the same name as the
    appropriate C struct (which must be defined in the C code that has been compiled to the ``ffi``
    object), *or* use an arbitrary name, but set the ``_name`` attribute to the C struct name.
    """

    _name = None
    _ffi = None

    def __init__(self):

        # Set the name of this struct in the C code
        if self._name is None:
            self._name = self.__class__.__name__

    @property
    def _cstruct(self):
        """
        The actual structure which needs to be passed around to C functions.

        .. note:: This is best accessed by calling the instance (see __call__).

        The reason it is defined as this (manual) cached property is so that it can be created
        dynamically, but not lost. It must not be lost, or else C functions which use it will lose
        access to its memory. But it also must be created dynamically so that it can be recreated
        after pickling (pickle can't handle CData).
        """
        try:
            return self.__cstruct
        except AttributeError:
            self.__cstruct = self._new()
            return self.__cstruct

    def _new(self):
        """Return a new empty C structure corresponding to this class."""
        return self._ffi.new("struct " + self._name + "*")

    @property
    def fields(self):
        """List of fields of the underlying C struct (a list of tuples of "name, type")."""
        return self._ffi.typeof(self._cstruct[0]).fields

    @property
    def fieldnames(self):
        """List names of fields of the underlying C struct."""
        return [f for f, t in self.fields]

    @property
    def pointer_fields(self):
        """List of names of fields which have pointer type in the C struct."""
        return [f for f, t in self.fields if t.type.kind == "pointer"]

    @property
    def primitive_fields(self):
        """List of names of fields which have primitive type in the C struct."""
        return [f for f, t in self.fields if t.type.kind == "primitive"]

    def __getstate__(self):
        """Return the current state of the class without pointers."""
        return {
            k: v
            for k, v in self.__dict__.items()
            if k not in ["_strings", "_StructWrapper__cstruct"]
        }

    def refresh_cstruct(self):
        """Delete the underlying C object, forcing it to be rebuilt."""
        try:
            del self.__cstruct
        except AttributeError:
            pass


class StructWithDefaults(StructWrapper):
    """
    A convenient interface to create a C structure with defaults specified.

    It is provided for the purpose of *creating* C structures in Python to be passed to C functions,
    where sensible defaults are available. Structures which are created within C and passed back do
    not need to be wrapped.

    This provides a *fully initialised* structure, and will fail if not all fields are specified
    with defaults.

    .. note:: The actual C structure is gotten by calling an instance. This is auto-generated when
              called, based on the parameters in the class.

    .. warning:: This class will *not* deal well with parameters of the struct which are pointers.
                 All parameters should be primitive types, except for strings, which are dealt with
                 specially.

    Parameters
    ----------
    ffi : cffi object
        The ffi object from any cffi-wrapped library.
    """

    _defaults_ = {}

    def __init__(self, *args, **kwargs):

        super().__init__()

        if args:
            if len(args) > 1:
                raise TypeError(
                    "%s takes up to one position argument, %s were given"
                    % (self.__class__.__name__, len(args))
                )
            elif args[0] is None:
                pass
            elif isinstance(args[0], self.__class__):
                kwargs.update(args[0].self)
            elif isinstance(args[0], dict):
                kwargs.update(args[0])
            else:
                raise TypeError(
                    "optional positional argument for %s must be None, dict, or an instance of itself"
                    % self.__class__.__name__
                )

        for k, v in self._defaults_.items():

            # Prefer arguments given to the constructor.
            _v = kwargs.pop(k, None)

            if _v is not None:
                v = _v

            try:
                setattr(self, k, v)
            except AttributeError:
                # The attribute has been defined as a property, save it as a hidden variable
                setattr(self, "_" + k, v)

        if kwargs:
            logger.warning(
                "The following parameters to {thisclass} are not supported: {lst}".format(
                    thisclass=self.__class__.__name__, lst=list(kwargs.keys())
                )
            )

    def convert(self, key, val):
        """Make any conversions of values before saving to the instance."""
        return val

    def update(self, **kwargs):
        """
        Update the parameters of an existing class structure.

        This should always be used instead of attempting to *assign* values to instance attributes.
        It consistently re-generates the underlying C memory space and sets some book-keeping
        variables.

        Parameters
        ----------
        kwargs:
            Any argument that may be passed to the class constructor.
        """
        # Start a fresh cstruct.
        if kwargs:
            self.refresh_cstruct()

        for k in self._defaults_:
            # Prefer arguments given to the constructor.
            if k in kwargs:
                v = kwargs.pop(k)

                try:
                    setattr(self, k, v)
                except AttributeError:
                    # The attribute has been defined as a property, save it as a hidden variable
                    setattr(self, "_" + k, v)

        # Also ensure that parameters that are part of the class, but not the defaults, are set
        # this will fail if these parameters cannot be set for some reason, hence doing it
        # last.
        for k in list(kwargs.keys()):
            if hasattr(self, k):
                setattr(self, k, kwargs.pop(k))

        if kwargs:
            warnings.warn(
                "The following arguments to be updated are not compatible with this class: %s"
                % kwargs
            )

    def __call__(self):
        """Return a filled C Structure corresponding to this instance."""
        for key, val in self.pystruct.items():

            # Find the value of this key in the current class
            if isinstance(val, str):
                # If it is a string, need to convert it to C string ourselves.
                val = self.ffi.new("char[]", getattr(self, key).encode())

            try:
                setattr(self._cstruct, key, val)
            except TypeError:
                print("For key %s, value %s:" % (key, val))
                raise

        return self._cstruct

    @property
    def pystruct(self):
        """A pure-python dictionary representation of the corresponding C structure."""
        return {fld: self.convert(fld, getattr(self, fld)) for fld in self.fieldnames}

    @property
    def defining_dict(self):
        """
        Pure python dictionary representation of this class, as it would appear in C.

        .. note:: This is not the same as :attr:`pystruct`, as it omits all variables that don't
                  need to be passed to the constructor, but appear in the C struct (some can be
                  calculated dynamically based on the inputs). It is also not the same as
                  :attr:`self`, as it includes the 'converted' values for each variable, which are
                  those actually passed to the C code.
        """
        return {k: self.convert(k, getattr(self, k)) for k in self._defaults_}

    @property
    def self(self):
        """
        Dictionary which if passed to its own constructor will yield an identical copy.

        .. note:: This differs from :attr:`pystruct` and :attr:`defining_dict` in that it uses the
                  hidden variable value, if it exists, instead of the exposed one. This prevents
                  from, for example, passing a value which is 10**10**val (and recurring!).
        """
        # Try to first use the hidden variable before using the non-hidden variety.
        dct = {}
        for k in self._defaults_:
            if hasattr(self, "_" + k):
                dct[k] = getattr(self, "_" + k)
            else:
                dct[k] = getattr(self, k)

        return dct

    def __repr__(self):
        """Full unique representation of the instance."""
        return (
            self.__class__.__name__
            + "("
            + ", ".join(sorted(k + ":" + str(v) for k, v in self.defining_dict.items()))
            + ")"
        )

    def __eq__(self, other):
        """Check whether this instance is equal to another object (by checking the __repr__)."""
        return self.__repr__() == repr(other)

    def __hash__(self):
        """Generate a unique hsh for the instance."""
        return hash(self.__repr__())


class OutputStruct(StructWrapper):
    """Base class for any class that wraps a C struct meant to be output from a C function."""

    _fields_ = []
    _global_params = None
    _inputs = ["user_params", "cosmo_params", "_random_seed"]
    _filter_params = ["external_table_path"]

    _TYPEMAP = {"float32": "float *", "float64": "double *", "int32": "int *"}

    def __init__(self, *, random_seed=None, init=False, dummy=False, **kwargs):
        super().__init__()

        self.filled = False
        self._random_seed = random_seed

        for k in self._inputs:

            if not hasattr(self, k):
                try:
                    setattr(self, k, kwargs.pop(k))
                except KeyError:
                    raise KeyError(
                        "%s requires the keyword argument %s"
                        % (self.__class__.__name__, k)
                    )

        if kwargs:
            warnings.warn(
                "%s received the following unexpected arguments: %s"
                % (self.__class__.__name__, list(kwargs.keys()))
            )

        self.dummy = dummy

        if init:
            self._init_cstruct()

    def _init_arrays(self):  # pragma: nocover
        """Abstract base method for initializing any arrays that the structure has."""
        pass

    @property
    def random_seed(self):
        """The random seed for this particular instance."""
        if self._random_seed is None:
            self._random_seed = int(np.random.randint(1, int(1e12)))

        return self._random_seed

    @property
    def arrays_initialized(self):
        """Whether all necessary arrays are initialized.

        .. note:: This must be true before passing to a C function.
        """
        # This assumes that all pointer fields will be arrays...
        for k in self.pointer_fields:
            if not hasattr(self, k):
                return False
            elif getattr(self._cstruct, k) == self._ffi.NULL:
                return False
        return True

    def _init_cstruct(self):
        if not self.filled:
            self._init_arrays()

        for k in self.pointer_fields:
            setattr(self._cstruct, k, self._ary2buf(getattr(self, k)))
        for k in self.primitive_fields:
            try:
                setattr(self._cstruct, k, getattr(self, k))
            except AttributeError:
                pass

        if not self.arrays_initialized:
            raise AttributeError(
                "%s is ill-defined. It has not initialized all necessary arrays."
                % self.__class__.__name__
            )

    def _ary2buf(self, ary):
        if not isinstance(ary, np.ndarray):
            raise ValueError("ary must be a numpy array")
        return self._ffi.cast(
            OutputStruct._TYPEMAP[ary.dtype.name], self._ffi.from_buffer(ary)
        )

    def __call__(self):
        """Initialize/allocate a fresh C struct in memory and return it."""
        if not self.arrays_initialized and not self.dummy:
            self._init_cstruct()

        return self._cstruct

    def _expose(self):
        """Expose the non-array primitives of the ctype to the top-level object."""
        if not self.filled:
            raise Exception(
                "You need to have actually called the C code before the primitives can be exposed."
            )
        for k in self.primitive_fields:
            setattr(self, k, getattr(self._cstruct, k))

    @property
    def _fname_skeleton(self):
        """The filename without specifying the random seed."""
        return self._name + "_" + self._md5 + "_r{seed}.h5"

    @property
    def filename(self):
        """The base filename of this object."""
        if self._random_seed is None:
            raise AttributeError("filename not defined until random_seed has been set")

        return self._fname_skeleton.format(seed=self.random_seed)

    def _get_fname(self, direc=None):
        direc = path.expanduser(direc or config["direc"])
        return path.join(direc, self.filename)

    def _find_file_without_seed(self, direc):
        allfiles = glob.glob(path.join(direc, self._fname_skeleton.format(seed="*")))
        if allfiles:
            return allfiles[0]
        else:
            return None

    def find_existing(self, direc=None):
        """
        Try to find existing boxes which match the parameters of this instance.

        Parameters
        ----------
        direc : str, optional
            The directory in which to search for the boxes. By default, this is the
            centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast/``.

        Returns
        -------
        str
            The filename of an existing set of boxes, or None.
        """
        # First, if appropriate, find a file without specifying seed.
        # Need to do this first, otherwise the seed will be chosen randomly upon
        # choosing a filename!
        direc = path.expanduser(direc or config["direc"])

        if not self._random_seed:
            f = self._find_file_without_seed(direc)
            if f and self._check_parameters(f):
                return f

        else:
            f = self._get_fname(direc)
            if path.exists(f) and self._check_parameters(f):
                return f
        return None

    def _check_parameters(self, fname):
        with h5py.File(fname, "r") as f:
            for k in self._inputs + ["_global_params"]:
                q = getattr(self, k)

                # The key name as it should appear in file.
                kfile = k.lstrip("_")

                # If this particular variable is set to None, this is interpreted
                # as meaning that we don't care about matching it to file.
                if q is None:
                    continue

                if isinstance(q, StructWithDefaults) or isinstance(
                    q, StructInstanceWrapper
                ):
                    grp = f[kfile]

                    if isinstance(q, StructWithDefaults):
                        dct = q.self
                    else:
                        dct = q

                    for kk, v in dct.items():
                        if kk not in self._filter_params:
                            file_v = grp.attrs[kk]
                            if file_v == "none":
                                file_v = None
                            if file_v != v:
                                logger.debug("For file %s:" % fname)
                                logger.debug(
                                    "\tThough md5 and seed matched, the parameter %s did not match, with values %s and %s in file and user respectively"
                                    % (kk, file_v, v)
                                )
                                return False
                else:
                    if f.attrs[kfile] != q:
                        return False
        return True

    def exists(self, direc=None):
        """
        Return a bool indicating whether a box matching the parameters of this instance is in cache.

        Parameters
        ----------
        direc : str, optional
            The directory in which to search for the boxes. By default, this is the
            centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast/``.
        """
        return self.find_existing(direc) is not None

    def write(self, direc=None, fname=None):
        """
        Write the struct in standard HDF5 format.

        Parameters
        ----------
        direc : str, optional
            The directory in which to search for the boxes. By default, this is the
            centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast/``.
        fname : str, optional
            The filename to write to. By default creates a unique filename from the hash.
        """
        if not self.filled:
            raise IOError("The boxes have not yet been computed.")

        if not self._random_seed:
            raise ValueError(
                "Attempting to write when no random seed has been set. Struct has been 'filled' inconsistently."
            )

        try:
            direc = path.expanduser(direc or config["direc"])
            fname = path.join(direc, fname) if fname else self._get_fname(direc)
            with h5py.File(fname, "w") as f:
                # Save input parameters to the file
                for k in self._inputs + ["_global_params"]:
                    q = getattr(self, k)

                    kfile = k.lstrip("_")

                    if isinstance(q, StructWithDefaults) or isinstance(
                        q, StructInstanceWrapper
                    ):
                        grp = f.create_group(kfile)
                        if isinstance(q, StructWithDefaults):
                            dct = (
                                q.self
                            )  # using self allows to rebuild the object from HDF5 file.
                        else:
                            dct = q

                        for kk, v in dct.items():
                            if kk not in self._filter_params:
                                grp.attrs[kk] = "none" if v is None else v
                    else:
                        f.attrs[kfile] = q

                # Save the boxes to the file
                boxes = f.create_group(self._name)

                # Go through all fields in this struct, and save
                for k in self.pointer_fields:
                    boxes.create_dataset(k, data=getattr(self, k))

                for k in self.primitive_fields:
                    boxes.attrs[k] = getattr(self, k)
        except OSError as e:
            logger.warning(
                "When attempting to write {} to file, write failed with the following error. Continuing without caching."
            )
            logger.warning(e)

    def read(self, direc=None):
        """
        Try find and read existing boxes from cache, which match the parameters of this instance.

        Parameters
        ----------
        direc : str, optional
            The directory in which to search for the boxes. By default, this is the
            centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast/``.
        """
        if self.filled:
            raise IOError("This data is already filled, no need to read in.")

        pth = self.find_existing(direc)

        if pth is None:
            raise IOError("No boxes exist for these parameters.")

        # Need to make sure arrays are initialized before reading in data to them.
        if not self.arrays_initialized:
            self._init_cstruct()

        with h5py.File(pth, "r") as f:
            try:
                boxes = f[self._name]
            except KeyError:
                raise IOError(
                    "While trying to read in %s, the file exists, but does not have the correct structure."
                    % self._name
                )

            # Fill our arrays.
            for k in boxes.keys():
                getattr(self, k)[...] = boxes[k][...]

            for k in boxes.attrs.keys():
                setattr(self, k, boxes.attrs[k])

            # Need to make sure that the seed is set to the one that's read in.
            seed = f.attrs["random_seed"]
            self._random_seed = seed

        self.filled = True
        self._expose()

    def __repr__(self):
        """Return a fully unique representation of the instance."""
        # This is the class name and all parameters which belong to C-based input structs,
        # eg. InitialConditions(HII_DIM:100,SIGMA_8:0.8,...)
        return self._seedless_repr() + "_random_seed={}".format(self._random_seed)

    def _seedless_repr(self):
        # The same as __repr__ except without the seed.
        return (
            self._name
            + "("
            + "; ".join(
                [
                    repr(v)
                    if isinstance(v, StructWithDefaults)
                    else (
                        v.filtered_repr(self._filter_params)
                        if isinstance(v, StructInstanceWrapper)
                        else k.lstrip("_") + ":" + repr(v)
                    )
                    for k, v in [
                        (k, getattr(self, k))
                        for k in self._inputs + ["_global_params"]
                        if k != "_random_seed"
                    ]
                ]
            )
            + ")"
        )

    def __str__(self):
        """Return a human-readable representation of the instance."""
        # this is *not* a unique representation, and doesn't include global params.
        return (
            self._name
            + "("
            + ";\n\t".join(
                [
                    repr(v)
                    if isinstance(v, StructWithDefaults)
                    else k.lstrip("_") + ":" + repr(v)
                    for k, v in [(k, getattr(self, k)) for k in self._inputs]
                ]
            )
            + ")"
        )

    def __hash__(self):
        """Return a unique hsh for this instance, even global params and random seed."""
        return hash(repr(self))

    @property
    def _md5(self):
        """Return a unique hsh of the object, *not* taking into account the random seed."""
        return md5(self._seedless_repr().encode()).hexdigest()

    def __eq__(self, other):
        """Check equality with another object via its __repr__."""
        return repr(self) == repr(other)


class StructInstanceWrapper:
    """A wrapper for *instances* of C structs.

    This is as opposed to :class:`StructWrapper`, which is for the un-instantiated structs.

    Parameters
    ----------
    wrapped :
        The reference to the C object to wrap (contained in the ``cffi.lib`` object).
    ffi :
        The ``cffi.ffi`` object.
    """

    def __init__(self, wrapped, ffi):
        self._cobj = wrapped
        self._ffi = ffi

        for nm, tp in self._ffi.typeof(self._cobj).fields:
            setattr(self, nm, getattr(self._cobj, nm))

        # Get the name of the structure
        self._ctype = self._ffi.typeof(self._cobj).cname.split()[-1]

    def __setattr__(self, name, value):
        """Set an attribute of the instance, attempting to change it in the C struct as well."""
        try:
            setattr(self._cobj, name, value)
        except AttributeError:
            pass
        object.__setattr__(self, name, value)

    def items(self):
        """Yield (name, value) pairs for each element of the struct."""
        for nm, tp in self._ffi.typeof(self._cobj).fields:
            yield nm, getattr(self, nm)

    def keys(self):
        """Return a list of names of elements in the struct."""
        return [nm for nm, tp in self.items()]

    def __repr__(self):
        """Return a unique representation of the instance."""
        return (
            self._ctype
            + "("
            + ";".join([k + "=" + str(v) for k, v in sorted(self.items())])
            + ")"
        )

    def filtered_repr(self, filter_params):
        """Get a fully unique representation of the instance that filters out some parametes.

        Parameters
        ----------
        filter_params : list of str
            The parameter names which should not appear in the representation.
        """
        return (
            self._ctype
            + "("
            + ";".join(
                [
                    k + "=" + str(v)
                    for k, v in sorted(self.items())
                    if k not in filter_params
                ]
            )
            + ")"
        )



import py21cmfast as p21c


def test_first_box():
    """Tests whether the first_box idea works for spin_temp.
    This test was breaking before we set the z_heat_max box to actually get
    the correct dimensions (before it was treated as a dummy).
    """
    initial_conditions = p21c.initial_conditions(
        user_params=p21c.UserParams(HII_DIM=100, BOX_LEN=100, DIM=200),
        random_seed=1,
        write=False,
    )

    perturbed_field = p21c.perturb_field(
        redshift=29.0, init_boxes=initial_conditions, write=False
    )

    spin_temp = p21c.spin_temperature(
        perturbed_field=perturbed_field, z_heat_max=30.0, write=False,
    )

    assert spin_temp.redshift == 29.0



"""
The main wrapper for the underlying 21cmFAST C-code.

The module provides both low- and high-level wrappers, using the very low-level machinery
in :mod:`~py21cmfast._utils`, and the convenient input and output structures from
:mod:`~py21cmfast.inputs` and :mod:`~py21cmfast.outputs`.

This module provides a number of:

* Low-level functions which simplify calling the background C functions which populate
  these output objects given the input classes.
* High-level functions which provide the most efficient and simplest way to generate the
  most commonly desired outputs.

**Low-level functions**

The low-level functions provided here ease the production of the aforementioned output
objects. Functions exist for each low-level C routine, which have been decoupled as far
as possible. So, functions exist to create :func:`initial_conditions`,
:func:`perturb_field`, :class:`ionize_box` and so on. Creating a brightness temperature
box (often the desired final output) would generally require calling each of these in
turn, as each depends on the result of a previous function. Nevertheless, each function
has the capability of generating the required previous outputs on-the-fly, so one can
instantly call :func:`ionize_box` and get a self-consistent result. Doing so, while
convenient, is sometimes not *efficient*, especially when using inhomogeneous
recombinations or the spin temperature field, which intrinsically require consistent
evolution of the ionization field through redshift. In these cases, for best efficiency
it is recommended to either use a customised manual approach to calling these low-level
functions, or to call a higher-level function which optimizes this process.

Finally, note that :mod:`py21cmfast` attempts to optimize the production of the large amount of
data via on-disk caching. By default, if a previous set of data has been computed using
the current input parameters, it will be read-in from a caching repository and returned
directly. This behaviour can be tuned in any of the low-level (or high-level) functions
by setting the `write`, `direc`, `regenerate` and `match_seed` parameters (see docs for
:func:`initial_conditions` for details). The function :func:`~query_cache` can be used
to search the cache, and return empty datasets corresponding to each (these can the be
filled with the data merely by calling ``.read()`` on any data set). Conversely, a
specific data set can be read and returned as a proper output object by calling the
:func:`~py21cmfast.cache_tools.readbox` function.


**High-level functions**

As previously mentioned, calling the low-level functions in some cases is non-optimal,
especially when full evolution of the field is required, and thus iteration through a
series of redshift. In addition, while :class:`InitialConditions` and
:class:`PerturbedField` are necessary intermediate data, it is *usually* the resulting
brightness temperature which is of most interest, and it is easier to not have to worry
about the intermediate steps explicitly. For these typical use-cases, two high-level
functions are available: :func:`run_coeval` and :func:`run_lightcone`, whose purpose
should be self-explanatory. These will optimally run all necessary intermediate
steps (using cached results by default if possible) and return all datasets of interest.


Examples
--------
A typical example of using this module would be the following.

>>> import py21cmfast as p21

Get coeval cubes at redshift 7,8 and 9, without spin temperature or inhomogeneous
recombinations:

>>> coeval = p21.run_coeval(
>>>     redshift=[7,8,9],
>>>     cosmo_params=p21.CosmoParams(hlittle=0.7),
>>>     user_params=p21.UserParams(HII_DIM=100)
>>> )

Get coeval cubes at the same redshift, with both spin temperature and inhomogeneous
recombinations, pulled from the natural evolution of the fields:

>>> all_boxes = p21.run_coeval(
>>>                 redshift=[7,8,9],
>>>                 user_params=p21.UserParams(HII_DIM=100),
>>>                 flag_options=p21.FlagOptions(INHOMO_RECO=True),
>>>                 do_spin_temp=True
>>>             )

Get a self-consistent lightcone defined between z1 and z2 (`z_step_factor` changes the
logarithmic steps between redshift that are actually evaluated, which are then
interpolated onto the lightcone cells):

>>> lightcone = p21.run_lightcone(redshift=z2, max_redshift=z2, z_step_factor=1.03)
"""
import logging
import os
from copy import deepcopy

import numpy as np
from astropy import units
from astropy.cosmology import z_at_value

from ._cfg import config
from ._utils import StructWrapper
from .c_21cmfast import ffi
from .c_21cmfast import lib
from .inputs import AstroParams
from .inputs import CosmoParams
from .inputs import FlagOptions
from .inputs import UserParams
from .inputs import global_params
from .outputs import BrightnessTemp
from .outputs import InitialConditions
from .outputs import IonizedBox
from .outputs import PerturbedField
from .outputs import TsBox

logger = logging.getLogger("21cmFAST")


def _check_compatible_inputs(*datasets, ignore=["redshift"]):
    """Ensure that all defined input parameters for the provided datasets are equal.

    Parameters
    ----------
    datasets : list of :class:`~_utils.OutputStruct`
        A number of output datasets to cross-check.
    ignore : list of str
        Attributes to ignore when ensuring that parameter inputs are the same.

    Raises
    ------
    ValueError :
        If datasets are not compatible.
    """
    done = []  # keeps track of inputs we've checked so we don't double check.

    for i, d in enumerate(datasets):
        # If a dataset is None, just ignore and move on.
        if d is None:
            continue

        # noinspection PyProtectedMember
        for inp in d._inputs:
            # Skip inputs that we want to ignore
            if inp in ignore:
                continue

            if inp not in done:
                for j, d2 in enumerate(datasets[(i + 1) :]):
                    if d2 is None:
                        continue

                    # noinspection PyProtectedMember
                    if inp in d2._inputs and getattr(d, inp) != getattr(d2, inp):
                        raise ValueError(
                            "%s and %s are incompatible"
                            % (d.__class__.__name__, d2.__class__.__name__)
                        )
                done += [inp]


def _configure_inputs(
    defaults: list,
    *datasets,
    ignore: list = ["redshift"],
    flag_none: [list, None] = None,
):
    """Configure a set of input parameter structs.

    This is useful for basing parameters on a previous output.
    The logic is this: the struct _cannot_ be present and different in both defaults and
    a dataset. If it is present in _either_ of them, that will be returned. If it is
    present in _neither_, either an error will be raised (if that item is in `flag_none`)
    or it will pass.

    Parameters
    ----------
    defaults : list of 2-tuples
        Each tuple is (key, val). Keys are input struct names, and values are a default
        structure for that input.
    datasets : list of :class:`~_utils.OutputStruct`
        A number of output datasets to cross-check, and draw parameter values from.
    ignore : list of str
        Attributes to ignore when ensuring that parameter inputs are the same.
    flag_none : list
        A list of parameter names for which ``None`` is not an acceptable value.

    Raises
    ------
    ValueError :
        If an input parameter is present in both defaults and the dataset, and is different.
        OR if the parameter is present in neither defaults not the datasets, and it is
        included in `flag_none`.
    """
    # First ensure all inputs are compaible in their parameters
    _check_compatible_inputs(*datasets, ignore=ignore)

    if flag_none is None:
        flag_none = []

    output = [0] * len(defaults)
    for i, (key, val) in enumerate(defaults):
        # Get the value of this input from the datasets
        data_val = None
        for dataset in datasets:
            if dataset is not None and hasattr(dataset, key):
                data_val = getattr(dataset, key)
                break

        # If both data and default have values
        if val is not None and data_val is not None and data_val != val:
            raise ValueError(
                "%s has an inconsistent value with %s"
                % (key, dataset.__class__.__name__)
            )
        else:
            if val is not None:
                output[i] = val
            elif data_val is not None:
                output[i] = data_val
            elif key in flag_none:
                raise ValueError(
                    "For %s, a value must be provided in some manner" % key
                )
            else:
                output[i] = None

    return output


def configure_redshift(redshift, *structs):
    """
    Check and obtain a redshift from given default and structs.

    Parameters
    ----------
    redshift : float
        The default redshift to use
    structs : list of :class:`~_utils.OutputStruct`
        A number of output datasets from which to find the redshift.

    Raises
    ------
    ValueError :
        If both `redshift` and *all* structs have a value of `None`, **or** if any of them
        are different from each other (and not `None`).
    """
    zs = set()
    for s in structs:
        if s is not None and hasattr(s, "redshift"):
            zs.add(s.redshift)

    zs = list(zs)

    if len(zs) > 1 or (len(zs) == 1 and redshift is not None and zs[0] != redshift):
        raise ValueError("Incompatible redshifts in inputs")
    elif len(zs) == 1:
        return zs[0]
    elif redshift is None:
        raise ValueError(
            "Either redshift must be provided, or a data set containing it."
        )
    else:
        return redshift


def _verify_types(**kwargs):
    """Ensure each argument has a type of None or that matching its name."""
    for k, v in kwargs.items():
        for j, kk in enumerate(["init", "perturb", "ionize", "spin_temp"]):
            if kk in k:
                break
        cls = [InitialConditions, PerturbedField, IonizedBox, TsBox][j]

        if v is not None and not isinstance(v, cls):
            raise ValueError("%s must be an instance of %s" % (k, cls.__name__))


class ParameterError(RuntimeError):
    """An exception representing a bad choice of parameters."""

    def __init__(self):
        default_message = "21cmFAST does not support this combination of parameters."
        super().__init__(default_message)


class FatalCError(Exception):
    """An exception representing something going wrong in C."""

    def __init__(self):
        default_message = "21cmFAST is exiting."
        super().__init__(default_message)


def _process_exitcode(exitcode):
    """Determine what happens for different values of the (integer) exit code from a C function."""
    if exitcode == 0:
        pass
    elif exitcode == 1:
        raise ParameterError
    elif exitcode == 2:
        raise FatalCError


def _call_c_func(fnc, obj, direc, *args, write=True):
    logger.debug("Calling {} with args: {}".format(fnc.__name__, args))
    exitcode = fnc(
        *[arg() if isinstance(arg, StructWrapper) else arg for arg in args], obj()
    )

    _process_exitcode(exitcode)
    obj.filled = True
    obj._expose()

    # Optionally do stuff with the result (like writing it)
    if write:
        obj.write(direc)

    return obj


def _get_config_options(direc, regenerate, write):
    return (
        os.path.expanduser(config["direc"] if direc is None else direc),
        config["regenerate"] if regenerate is None else regenerate,
        config["write"] if write is None else write,
    )


# ======================================================================================
# WRAPPING FUNCTIONS
# ======================================================================================
def compute_tau(*, redshifts, global_xHI, user_params=None, cosmo_params=None):
    """Compute the optical depth to reionization under the given model.

    Parameters
    ----------
    redshifts : array-like
        Redshifts defining an evolution of the neutral fraction.
    global_xHI : array-like
        The mean neutral fraction at `redshifts`.
    user_params : :class:`~inputs.UserParams`
        Parameters defining the simulation run.
    cosmo_params : :class:`~inputs.CosmoParams`
        Cosmological parameters.

    Returns
    -------
    tau : float
        The optional depth to reionization

    Raises
    ------
    ValueError :
        If `redshifts` and `global_xHI` have inconsistent length.
    """
    user_params = UserParams(user_params)
    cosmo_params = CosmoParams(cosmo_params)

    if len(redshifts) != len(global_xHI):
        raise ValueError("redshifts and global_xHI must have same length")

    # Convert the data to the right type
    redshifts = np.array(redshifts, dtype="float32")
    global_xHI = np.array(global_xHI, dtype="float32")

    z = ffi.cast("float *", ffi.from_buffer(redshifts))
    xHI = ffi.cast("float *", ffi.from_buffer(global_xHI))

    # Run the C code
    return lib.ComputeTau(user_params(), cosmo_params(), len(redshifts), z, xHI)


def compute_luminosity_function(
    *,
    redshifts,
    user_params=None,
    cosmo_params=None,
    astro_params=None,
    flag_options=None,
    nbins=100,
):
    """Compute a the luminosity function over a given number of bins and redshifts.

    Parameters
    ----------
    redshifts : array-like
        The redshifts at which to compute the luminosity function.
    user_params : :class:`~UserParams`, optional
        Defines the overall options and parameters of the run.
    cosmo_params : :class:`~CosmoParams`, optional
        Defines the cosmological parameters used to compute initial conditions.
    astro_params : :class:`~AstroParams`, optional
        The astrophysical parameters defining the course of reionization.
    flag_options : :class:`~FlagOptions`, optional
        Some options passed to the reionization routine.
    nbins : int, optional
        The number of luminosity bins to produce for the luminosity function.

    Returns
    -------
    Muvfunc : np.ndarray
        Magnitude array (i.e. brightness). Shape [nredshifts, nbins]
    Mhfunc : np.ndarray
        Halo mass array. Shape [nredshifts, nbins]
    lfunc : np.ndarray
        Number density of haloes corresponding to each bin defined by `Muvfunc`.
        Shape [nredshifts, nbins].
    """
    user_params = UserParams(user_params)
    cosmo_params = CosmoParams(cosmo_params)
    astro_params = AstroParams(astro_params)
    flag_options = FlagOptions(flag_options)

    redshifts = np.array(redshifts, dtype="float32")

    lfunc = np.zeros(len(redshifts) * nbins)
    Muvfunc = np.zeros(len(redshifts) * nbins)
    Mhfunc = np.zeros(len(redshifts) * nbins)

    lfunc.shape = (len(redshifts), nbins)
    Muvfunc.shape = (len(redshifts), nbins)
    Mhfunc.shape = (len(redshifts), nbins)

    c_Muvfunc = ffi.cast("double *", ffi.from_buffer(Muvfunc))
    c_Mhfunc = ffi.cast("double *", ffi.from_buffer(Mhfunc))
    c_lfunc = ffi.cast("double *", ffi.from_buffer(lfunc))

    # Run the C code
    errcode = lib.ComputeLF(
        nbins,
        user_params(),
        cosmo_params(),
        astro_params(),
        flag_options(),
        len(redshifts),
        ffi.cast("float *", ffi.from_buffer(redshifts)),
        c_Muvfunc,
        c_Mhfunc,
        c_lfunc,
    )

    _process_exitcode(errcode)

    lfunc[lfunc <= -30] = np.nan

    return Muvfunc, Mhfunc, lfunc


def _init_photon_conservation_correction(
    *, user_params=None, cosmo_params=None, astro_params=None, flag_options=None
):
    user_params = UserParams(user_params)
    cosmo_params = CosmoParams(cosmo_params)
    astro_params = AstroParams(astro_params)
    flag_options = FlagOptions(flag_options)

    return lib.InitialisePhotonCons(
        user_params(), cosmo_params(), astro_params(), flag_options()
    )


def _calibrate_photon_conservation_correction(
    *, redshifts_estimate, nf_estimate, NSpline
):
    # Convert the data to the right type
    redshifts_estimate = np.array(redshifts_estimate, dtype="float64")
    nf_estimate = np.array(nf_estimate, dtype="float64")

    z = ffi.cast("double *", ffi.from_buffer(redshifts_estimate))
    xHI = ffi.cast("double *", ffi.from_buffer(nf_estimate))

    return lib.PhotonCons_Calibration(z, xHI, NSpline)


def _calc_zstart_photon_cons():
    # Run the C code
    return lib.ComputeZstart_PhotonCons()


def _get_photon_nonconservation_data():
    """
    Access C global data representing the photon-nonconservation corrections.

    .. note::  if not using ``PHOTON_CONS`` (in :class:`~FlagOptions`), *or* if the
               initialisation for photon conservation has not been performed yet, this
               will return None.

    Returns
    -------
    dict :
      z_analytic: array of redshifts defining the analytic ionized fraction
      Q_analytic: array of analytic  ionized fractions corresponding to `z_analytic`
      z_calibration: array of redshifts defining the ionized fraction from 21cmFAST without
      recombinations
      nf_calibration: array of calibration ionized fractions corresponding to `z_calibration`
      delta_z_photon_cons: the change in redshift required to calibrate 21cmFAST, as a function
      of z_calibration
      nf_photoncons: the neutral fraction as a function of redshift
    """
    # Check if photon conservation has been initialised at all
    if not lib.photon_cons_inited:
        return None

    arbitrary_large_size = 2000

    data = np.zeros((6, arbitrary_large_size))

    IntVal1 = np.array(np.zeros(1), dtype="int32")
    IntVal2 = np.array(np.zeros(1), dtype="int32")
    IntVal3 = np.array(np.zeros(1), dtype="int32")

    c_z_at_Q = ffi.cast("double *", ffi.from_buffer(data[0]))
    c_Qval = ffi.cast("double *", ffi.from_buffer(data[1]))
    c_z_cal = ffi.cast("double *", ffi.from_buffer(data[2]))
    c_nf_cal = ffi.cast("double *", ffi.from_buffer(data[3]))
    c_PC_nf = ffi.cast("double *", ffi.from_buffer(data[4]))
    c_PC_deltaz = ffi.cast("double *", ffi.from_buffer(data[5]))

    c_int_NQ = ffi.cast("int *", ffi.from_buffer(IntVal1))
    c_int_NC = ffi.cast("int *", ffi.from_buffer(IntVal2))
    c_int_NP = ffi.cast("int *", ffi.from_buffer(IntVal3))

    # Run the C code
    errcode = lib.ObtainPhotonConsData(
        c_z_at_Q,
        c_Qval,
        c_int_NQ,
        c_z_cal,
        c_nf_cal,
        c_int_NC,
        c_PC_nf,
        c_PC_deltaz,
        c_int_NP,
    )

    _process_exitcode(errcode)

    ArrayIndices = [
        IntVal1[0],
        IntVal1[0],
        IntVal2[0],
        IntVal2[0],
        IntVal3[0],
        IntVal3[0],
    ]

    data_list = [
        "z_analytic",
        "Q_analytic",
        "z_calibration",
        "nf_calibration",
        "delta_z_photon_cons",
        "nf_photoncons",
    ]

    photon_nonconservation_data = {
        name: d[:index] for name, d, index in zip(data_list, data, ArrayIndices)
    }

    return photon_nonconservation_data


def initial_conditions(
    *,
    user_params=None,
    cosmo_params=None,
    random_seed=None,
    regenerate=None,
    write=None,
    direc=None,
    **global_kwargs,
):
    r"""
    Compute initial conditions.

    Parameters
    ----------
    user_params : :class:`~UserParams` instance, optional
        Defines the overall options and parameters of the run.
    cosmo_params : :class:`~CosmoParams` instance, optional
        Defines the cosmological parameters used to compute initial conditions.
    regenerate : bool, optional
        Whether to force regeneration of data, even if matching cached data is found.
        This is applied recursively to any potential sub-calculations. It is ignored in
        the case of dependent data only if that data is explicitly passed to the function.
    write : bool, optional
        Whether to write results to file (i.e. cache). This is recursively applied to
        any potential sub-calculations.
    direc : str, optional
        The directory in which to search for the boxes and write them. By default, this
        is the directory given by ``boxdir`` in the configuration file,
        ``~/.21cmfast/config.yml``. This is recursively applied to any potential
        sub-calculations.
    \*\*global_kwargs :
        Any attributes for :class:`~py21cmfast.inputs.GlobalParams`. This will
        *temporarily* set global attributes for the duration of the function. Note that
        arguments will be treated as case-insensitive.

    Returns
    -------
    :class:`~InitialConditions`
    """
    direc, regenerate, write = _get_config_options(direc, regenerate, write)

    with global_params.use(**global_kwargs):
        user_params = UserParams(user_params)
        cosmo_params = CosmoParams(cosmo_params)

        # Initialize memory for the boxes that will be returned.
        boxes = InitialConditions(
            user_params=user_params, cosmo_params=cosmo_params, random_seed=random_seed
        )

        # First check whether the boxes already exist.
        if not regenerate:
            try:
                boxes.read(direc)
                logger.info(
                    "Existing init_boxes found and read in (seed=%s)."
                    % boxes.random_seed
                )
                return boxes
            except IOError:
                pass

        return _call_c_func(
            lib.ComputeInitialConditions,
            boxes,
            direc,
            boxes.random_seed,
            boxes.user_params,
            boxes.cosmo_params,
            write=write,
        )


def perturb_field(
    *,
    redshift,
    init_boxes=None,
    user_params=None,
    cosmo_params=None,
    random_seed=None,
    regenerate=None,
    write=None,
    direc=None,
    **global_kwargs,
):
    r"""
    Compute a perturbed field at a given redshift.

    Parameters
    ----------
    redshift : float
        The redshift at which to compute the perturbed field.
    init_boxes : :class:`~InitialConditions`, optional
        If given, these initial conditions boxes will be used, otherwise initial conditions will
        be generated. If given,
        the user and cosmo params will be set from this object.
    user_params : :class:`~UserParams`, optional
        Defines the overall options and parameters of the run.
    cosmo_params : :class:`~CosmoParams`, optional
        Defines the cosmological parameters used to compute initial conditions.
    \*\*global_kwargs :
        Any attributes for :class:`~py21cmfast.inputs.GlobalParams`. This will
        *temporarily* set global attributes for the duration of the function. Note that
        arguments will be treated as case-insensitive.

    Returns
    -------
    :class:`~PerturbedField`

    Other Parameters
    ----------------
    regenerate, write, direc, random_seed:
        See docs of :func:`initial_conditions` for more information.

    Examples
    --------
    The simplest method is just to give a redshift::

    >>> field = perturb_field(7.0)
    >>> print(field.density)

    Doing so will internally call the :func:`~initial_conditions` function. If initial conditions
    have already been
    calculated, this can be avoided by passing them:

    >>> init_boxes = initial_conditions()
    >>> field7 = perturb_field(7.0, init_boxes)
    >>> field8 = perturb_field(8.0, init_boxes)

    The user and cosmo parameter structures are by default inferred from the ``init_boxes``,
    so that the following is
    consistent::

    >>> init_boxes = initial_conditions(user_params= UserParams(HII_DIM=1000))
    >>> field7 = perturb_field(7.0, init_boxes)

    If ``init_boxes`` is not passed, then these parameters can be directly passed::

    >>> field7 = perturb_field(7.0, user_params=UserParams(HII_DIM=1000))

    """
    direc, regenerate, write = _get_config_options(direc, regenerate, write)
    with global_params.use(**global_kwargs):
        _verify_types(init_boxes=init_boxes)

        # Configure and check input/output parameters/structs
        random_seed, user_params, cosmo_params = _configure_inputs(
            [
                ("random_seed", random_seed),
                ("user_params", user_params),
                ("cosmo_params", cosmo_params),
            ],
            init_boxes,
        )

        # Verify input parameter structs (need to do this after configure_inputs).
        user_params = UserParams(user_params)
        cosmo_params = CosmoParams(cosmo_params)

        # Initialize perturbed boxes.
        fields = PerturbedField(
            redshift=redshift,
            user_params=user_params,
            cosmo_params=cosmo_params,
            random_seed=random_seed,
        )

        # Check whether the boxes already exist
        if not regenerate:
            try:
                fields.read(direc)
                logger.info(
                    "Existing z=%s perturb_field boxes found and read in (seed=%s)."
                    % (redshift, fields.random_seed)
                )
                return fields
            except IOError:
                pass

        # Make sure we've got computed init boxes.
        if init_boxes is None or not init_boxes.filled:
            init_boxes = initial_conditions(
                user_params=user_params,
                cosmo_params=cosmo_params,
                regenerate=regenerate,
                write=write,
                direc=direc,
                random_seed=random_seed,
            )

            # Need to update fields to have the same seed as init_boxes
            fields._random_seed = init_boxes.random_seed

        # Run the C Code
        return _call_c_func(
            lib.ComputePerturbField,
            fields,
            direc,
            redshift,
            fields.user_params,
            fields.cosmo_params,
            init_boxes,
            write=write,
        )


def ionize_box(
    *,
    astro_params=None,
    flag_options=None,
    redshift=None,
    perturbed_field=None,
    previous_ionize_box=None,
    spin_temp=None,
    init_boxes=None,
    cosmo_params=None,
    user_params=None,
    regenerate=None,
    write=None,
    direc=None,
    random_seed=None,
    cleanup=True,
    **global_kwargs,
):
    r"""
    Compute an ionized box at a given redshift.

    This function has various options for how the evolution of the ionization is computed (if at
    all). See the Notes
    below for details.

    Parameters
    ----------
    astro_params: :class:`~AstroParams` instance, optional
        The astrophysical parameters defining the course of reionization.
    flag_options: :class:`~FlagOptions` instance, optional
        Some options passed to the reionization routine.
    redshift : float, optional
        The redshift at which to compute the ionized box. If `perturbed_field` is given,
        its inherent redshift
        will take precedence over this argument. If not, this argument is mandatory.
    perturbed_field : :class:`~PerturbField`, optional
        If given, this field will be used, otherwise it will be generated. To be generated,
        either `init_boxes` and
        `redshift` must be given, or `user_params`, `cosmo_params` and `redshift`.
    init_boxes : :class:`~InitialConditions` , optional
        If given, and `perturbed_field` *not* given, these initial conditions boxes will be used
        to generate the perturbed field, otherwise initial conditions will be generated on the fly.
        If given, the user and cosmo params will be set from this object.
    previous_ionize_box: :class:`IonizedBox` or None
        An ionized box at higher redshift. This is only used if `INHOMO_RECO` and/or `do_spin_temp`
        are true. If either of these are true, and this is not given, then it will be assumed that
        this is the "first box", i.e. that it can be populated accurately without knowing source
        statistics.
    spin_temp: :class:`TsBox` or None, optional
        A spin-temperature box, only required if `do_spin_temp` is True. If None, will try to read
        in a spin temp box at the current redshift, and failing that will try to automatically
        create one, using the previous ionized box redshift as the previous spin temperature
        redshift.
    user_params : :class:`~UserParams`, optional
        Defines the overall options and parameters of the run.
    cosmo_params : :class:`~CosmoParams`, optional
        Defines the cosmological parameters used to compute initial conditions.
    cleanup : bool, optional
        A flag to specify whether the C routine cleans up its memory before returning. Typically,
        if `spin_temperature` is called directly, you will want this to be true, as if the next box
        to be calculate has different shape, errors will occur if memory is not cleaned. However,
        it can be useful to set it to False if scrolling through parameters for the same box shape.
    \*\*global_kwargs :
        Any attributes for :class:`~py21cmfast.inputs.GlobalParams`. This will
        *temporarily* set global attributes for the duration of the function. Note that
        arguments will be treated as case-insensitive.

    Returns
    -------
    :class:`~IonizedBox` :
        An object containing the ionized box data.

    Other Parameters
    ----------------
    regenerate, write, direc, random_seed :
        See docs of :func:`initial_conditions` for more information.

    Notes
    -----
    Typically, the ionization field at any redshift is dependent on the evolution of xHI up until
    that redshift, which necessitates providing a previous ionization field to define the current
    one. This function provides several options for doing so. First, if neither the spin
    temperature field, nor inhomogeneous recombinations (specified in flag options) are used, no
    evolution needs to be done. Otherwise, either (in order of precedence)

    1. a specific previous :class`~IonizedBox` object is provided, which will be used directly,
    2. a previous redshift is provided, for which a cached field on disk will be sought,
    3. a step factor is provided which recursively steps through redshift, calculating previous
       fields up until Z_HEAT_MAX, and returning just the final field at the current redshift, or
    4. the function is instructed to treat the current field as being an initial "high-redshift"
       field such that specific sources need not be found and evolved.

    .. note:: If a previous specific redshift is given, but no cached field is found at that
              redshift, the previous ionization field will be evaluated based on `z_step_factor`.

    Examples
    --------
    By default, no spin temperature is used, and neither are inhomogeneous recombinations,
    so that no evolution is required, thus the following will compute a coeval ionization box:

    >>> xHI = ionize_box(redshift=7.0)

    However, if either of those options are true, then a full evolution will be required:

    >>> xHI = ionize_box(redshift=7.0, flag_options=FlagOptions(INHOMO_RECO=True,USE_TS_FLUCT=True))

    This will by default evolve the field from a redshift of *at least* `Z_HEAT_MAX` (a global
    parameter), in logarithmic steps of `ZPRIME_STEP_FACTOR`. To change these:

    >>> xHI = ionize_box(redshift=7.0, zprime_step_factor=1.2, z_heat_max=15.0,
    >>>                  flag_options={"USE_TS_FLUCT":True})

    Alternatively, one can pass an exact previous redshift, which will be sought in the disk
    cache, or evaluated:

    >>> ts_box = ionize_box(redshift=7.0, previous_ionize_box=8.0, flag_options={
    >>>                     "USE_TS_FLUCT":True})

    Beware that doing this, if the previous box is not found on disk, will continue to evaluate
    prior boxes based on `ZPRIME_STEP_FACTOR`. Alternatively, one can pass a previous
    :class:`~IonizedBox`:

    >>> xHI_0 = ionize_box(redshift=8.0, flag_options={"USE_TS_FLUCT":True})
    >>> xHI = ionize_box(redshift=7.0, previous_ionize_box=xHI_0)

    Again, the first line here will implicitly use ``ZPRIME_STEP_FACTOR`` to evolve the field from
    ``Z_HEAT_MAX``. Note that in the second line, all of the input parameters are taken directly from
    `xHI_0` so that they are consistent, and we need not specify the ``flag_options``.

    As the function recursively evaluates previous redshift, the previous spin temperature fields
    will also be consistently recursively evaluated. Only the final ionized box will actually be
    returned and kept in memory, however intervening results will by default be cached on disk.
    One can also pass an explicit spin temperature object:

    >>> ts = spin_temperature(redshift=7.0)
    >>> xHI = ionize_box(redshift=7.0, spin_temp=ts)

    If automatic recursion is used, then it is done in such a way that no large boxes are kept
    around in memory for longer than they need to be (only two at a time are required).
    """
    direc, regenerate, write = _get_config_options(direc, regenerate, write)

    with global_params.use(**global_kwargs):
        _verify_types(
            init_boxes=init_boxes,
            perturbed_field=perturbed_field,
            previous_ionize_box=previous_ionize_box,
            spin_temp=spin_temp,
        )

        # Configure and check input/output parameters/structs
        (
            random_seed,
            user_params,
            cosmo_params,
            astro_params,
            flag_options,
        ) = _configure_inputs(
            [
                ("random_seed", random_seed),
                ("user_params", user_params),
                ("cosmo_params", cosmo_params),
                ("astro_params", astro_params),
                ("flag_options", flag_options),
            ],
            init_boxes,
            spin_temp,
            init_boxes,
            perturbed_field,
            previous_ionize_box,
        )
        redshift = configure_redshift(redshift, spin_temp, perturbed_field)

        # Verify input structs
        user_params = UserParams(user_params)
        cosmo_params = CosmoParams(cosmo_params)
        flag_options = FlagOptions(flag_options)
        astro_params = AstroParams(astro_params, INHOMO_RECO=flag_options.INHOMO_RECO)

        if spin_temp is not None and not flag_options.USE_TS_FLUCT:
            logger.warning(
                "Changing flag_options.USE_TS_FLUCT to True since spin_temp was passed."
            )
            flag_options.USE_TS_FLUCT = True

        box = IonizedBox(
            first_box=((1 + redshift) * global_params.ZPRIME_STEP_FACTOR - 1)
            > global_params.Z_HEAT_MAX
            and (
                not isinstance(previous_ionize_box, IonizedBox)
                or not previous_ionize_box.filled
            ),
            user_params=user_params,
            cosmo_params=cosmo_params,
            redshift=redshift,
            astro_params=astro_params,
            flag_options=flag_options,
            random_seed=random_seed,
        )

        # Check whether the boxes already exist
        if not regenerate:
            try:
                box.read(direc)
                logger.info(
                    "Existing z=%s ionized boxes found and read in (seed=%s)."
                    % (redshift, box.random_seed)
                )
                return box
            except IOError:
                pass

        # EVERYTHING PAST THIS POINT ONLY HAPPENS IF THE BOX DOESN'T ALREADY EXIST
        # ------------------------------------------------------------------------
        # Get the previous redshift
        if flag_options.INHOMO_RECO or flag_options.USE_TS_FLUCT:

            if previous_ionize_box is not None:
                prev_z = previous_ionize_box.redshift
            else:
                prev_z = (1 + redshift) * global_params.ZPRIME_STEP_FACTOR - 1

            # Ensure the previous spin temperature has a higher redshift than this one.
            if prev_z and prev_z <= redshift:
                raise ValueError(
                    "Previous ionized box must have a higher redshift than that being evaluated."
                )
        else:
            prev_z = None

        # Get init_box required.
        if init_boxes is None or not init_boxes.filled:
            init_boxes = initial_conditions(
                user_params=user_params,
                cosmo_params=cosmo_params,
                regenerate=regenerate,
                write=write,
                direc=direc,
                random_seed=random_seed,
            )

            # Need to update random seed
            box._random_seed = init_boxes.random_seed

        # Get appropriate previous ionization box
        if previous_ionize_box is None or not previous_ionize_box.filled:
            # If we are beyond Z_HEAT_MAX, just make an empty box
            if prev_z is None or prev_z > global_params.Z_HEAT_MAX:
                previous_ionize_box = IonizedBox(redshift=0)

            # Otherwise recursively create new previous box.
            else:
                previous_ionize_box = ionize_box(
                    astro_params=astro_params,
                    flag_options=flag_options,
                    redshift=prev_z,
                    init_boxes=init_boxes,
                    regenerate=regenerate,
                    write=write,
                    direc=direc,
                    cleanup=False,  # We *know* we're going to need the memory again.
                )

        # Dynamically produce the perturbed field.
        if perturbed_field is None or not perturbed_field.filled:
            perturbed_field = perturb_field(
                init_boxes=init_boxes,
                # NOTE: this is required, rather than using cosmo_ and user_,
                # since init may have a set seed.
                redshift=redshift,
                regenerate=regenerate,
                write=write,
                direc=direc,
            )

        # Set empty spin temp box if necessary.
        if not flag_options.USE_TS_FLUCT:
            spin_temp = TsBox(redshift=0)
        elif spin_temp is None:
            spin_temp = spin_temperature(
                perturbed_field=perturbed_field,
                flag_options=flag_options,
                init_boxes=init_boxes,
                direc=direc,
                write=write,
                regenerate=regenerate,
                cleanup=cleanup,
            )

        # Run the C Code
        return _call_c_func(
            lib.ComputeIonizedBox,
            box,
            direc,
            redshift,
            previous_ionize_box.redshift,
            box.user_params,
            box.cosmo_params,
            box.astro_params,
            box.flag_options,
            perturbed_field,
            previous_ionize_box,
            spin_temp,
            write=write,
        )


def spin_temperature(
    *,
    astro_params=None,
    flag_options=None,
    redshift=None,
    perturbed_field=None,
    previous_spin_temp=None,
    init_boxes=None,
    cosmo_params=None,
    user_params=None,
    regenerate=None,
    write=None,
    direc=None,
    random_seed=None,
    cleanup=True,
    **global_kwargs,
):
    r"""
    Compute spin temperature boxes at a given redshift.

    See the notes below for how the spin temperature field is evolved through redshift.

    Parameters
    ----------
    astro_params : :class:`~AstroParams`, optional
        The astrophysical parameters defining the course of reionization.
    flag_options : :class:`~FlagOptions`, optional
        Some options passed to the reionization routine.
    redshift : float, optional
        The redshift at which to compute the ionized box. If not given, the redshift from
        `perturbed_field` will be used. Either `redshift`, `perturbed_field`, or
        `previous_spin_temp` must be given. See notes on `perturbed_field` for how it affects the
        given redshift if both are given.
    perturbed_field : :class:`~PerturbField`, optional
        If given, this field will be used, otherwise it will be generated. To be generated,
        either `init_boxes` and `redshift` must be given, or `user_params`, `cosmo_params` and
        `redshift`. By default, this will be generated at the same redshift as the spin temperature
        box. The redshift of perturb field is allowed to be different than `redshift`. If so, it
        will be interpolated to the correct redshift, which can provide a speedup compared to
        actually computing it at the desired redshift.
    previous_spin_temp : :class:`TsBox` or None
        The previous spin temperature box.
    init_boxes : :class:`~InitialConditions`, optional
        If given, and `perturbed_field` *not* given, these initial conditions boxes will be used
        to generate the perturbed field, otherwise initial conditions will be generated on the fly.
        If given, the user and cosmo params will be set from this object.
    user_params : :class:`~UserParams`, optional
        Defines the overall options and parameters of the run.
    cosmo_params : :class:`~CosmoParams`, optional
        Defines the cosmological parameters used to compute initial conditions.
    cleanup : bool, optional
        A flag to specify whether the C routine cleans up its memory before returning.
        Typically, if `spin_temperature` is called directly, you will want this to be
        true, as if the next box to be calculate has different shape, errors will occur
        if memory is not cleaned. However, it can be useful to set it to False if
        scrolling through parameters for the same box shape.
    \*\*global_kwargs :
        Any attributes for :class:`~py21cmfast.inputs.GlobalParams`. This will
        *temporarily* set global attributes for the duration of the function. Note that
        arguments will be treated as case-insensitive.

    Returns
    -------
    :class:`~TsBox`
        An object containing the spin temperature box data.

    Other Parameters
    ----------------
    regenerate, write, direc, random_seed :
        See docs of :func:`initial_conditions` for more information.

    Notes
    -----
    Typically, the spin temperature field at any redshift is dependent on the evolution of spin
    temperature up until that redshift, which necessitates providing a previous spin temperature
    field to define the current one. This function provides several options for doing so. Either
    (in order of precedence):

    1. a specific previous spin temperature object is provided, which will be used directly,
    2. a previous redshift is provided, for which a cached field on disk will be sought,
    3. a step factor is provided which recursively steps through redshift, calculating previous
       fields up until Z_HEAT_MAX, and returning just the final field at the current redshift, or
    4. the function is instructed to treat the current field as being an initial "high-redshift"
       field such that specific sources need not be found and evolved.

    .. note:: If a previous specific redshift is given, but no cached field is found at that
              redshift, the previous spin temperature field will be evaluated based on
              ``z_step_factor``.

    Examples
    --------
    To calculate and return a fully evolved spin temperature field at a given redshift (with
    default input parameters), simply use:

    >>> ts_box = spin_temperature(redshift=7.0)

    This will by default evolve the field from a redshift of *at least* `Z_HEAT_MAX` (a global
    parameter), in logarithmic steps of `z_step_factor`. Thus to change these:

    >>> ts_box = spin_temperature(redshift=7.0, zprime_step_factor=1.2, z_heat_max=15.0)

    Alternatively, one can pass an exact previous redshift, which will be sought in the disk
    cache, or evaluated:

    >>> ts_box = spin_temperature(redshift=7.0, previous_spin_temp=8.0)

    Beware that doing this, if the previous box is not found on disk, will continue to evaluate
    prior boxes based on the ``z_step_factor``. Alternatively, one can pass a previous spin
    temperature box:

    >>> ts_box1 = spin_temperature(redshift=8.0)
    >>> ts_box = spin_temperature(redshift=7.0, previous_spin_temp=ts_box1)

    Again, the first line here will implicitly use ``z_step_factor`` to evolve the field from
    around ``Z_HEAT_MAX``. Note that in the second line, all of the input parameters are taken
    directly from `ts_box1` so that they are consistent. Finally, one can force the function to
    evaluate the current redshift as if it was beyond ``Z_HEAT_MAX`` so that it depends only on
    itself:

    >>> ts_box = spin_temperature(redshift=7.0, zprime_step_factor=None)

    This is usually a bad idea, and will give a warning, but it is possible.
    """
    direc, regenerate, write = _get_config_options(direc, regenerate, write)

    with global_params.use(**global_kwargs):
        _verify_types(
            init_boxes=init_boxes,
            perturbed_field=perturbed_field,
            previous_spin_temp=previous_spin_temp,
        )

        # Configure and check input/output parameters/structs
        (
            random_seed,
            user_params,
            cosmo_params,
            astro_params,
            flag_options,
        ) = _configure_inputs(
            [
                ("random_seed", random_seed),
                ("user_params", user_params),
                ("cosmo_params", cosmo_params),
                ("astro_params", astro_params),
                ("flag_options", flag_options),
            ],
            init_boxes,
            previous_spin_temp,
            init_boxes,
            perturbed_field,
        )

        # Try to determine redshift from other inputs, if required.
        # Note that perturb_field does not need to match redshift here.
        if redshift is None:
            if perturbed_field is not None:
                redshift = perturbed_field.redshift
            elif previous_spin_temp is not None:
                redshift = (
                    previous_spin_temp.redshift + 1
                ) / global_params.ZPRIME_STEP_FACTOR - 1
            else:
                raise ValueError(
                    "Either the redshift, perturbed_field or previous_spin_temp must be given."
                )
        user_params = UserParams(user_params)
        cosmo_params = CosmoParams(cosmo_params)
        flag_options = FlagOptions(flag_options)
        astro_params = AstroParams(astro_params, INHOMO_RECO=flag_options.INHOMO_RECO)

        # Explicitly set this flag to True, though it shouldn't be required!
        flag_options.USE_TS_FLUCT = True

        box = TsBox(
            first_box=((1 + redshift) * global_params.ZPRIME_STEP_FACTOR - 1)
            > global_params.Z_HEAT_MAX
            and (
                not isinstance(previous_spin_temp, IonizedBox)
                or not previous_spin_temp.filled
            ),
            user_params=user_params,
            cosmo_params=cosmo_params,
            redshift=redshift,
            astro_params=astro_params,
            flag_options=flag_options,
            random_seed=random_seed,
        )

        # Check whether the boxes already exist on disk.
        if not regenerate:
            try:
                box.read(direc)
                logger.info(
                    "Existing z=%s spin_temp boxes found and read in (seed=%s)."
                    % (redshift, box.random_seed)
                )
                return box
            except IOError:
                pass

        # EVERYTHING PAST THIS POINT ONLY HAPPENS IF THE BOX DOESN'T ALREADY EXIST
        # ------------------------------------------------------------------------

        # Get the previous redshift
        if previous_spin_temp is not None:
            prev_z = previous_spin_temp.redshift
        else:
            prev_z = (1 + redshift) * global_params.ZPRIME_STEP_FACTOR - 1

        # Ensure the previous spin temperature has a higher redshift than this one.
        if prev_z and prev_z <= redshift:
            raise ValueError(
                "Previous spin temperature box must have a higher redshift than that being evaluated."
            )

        # Dynamically produce the initial conditions.
        if init_boxes is None or not init_boxes.filled:
            init_boxes = initial_conditions(
                user_params=user_params,
                cosmo_params=cosmo_params,
                regenerate=regenerate,
                write=write,
                direc=direc,
                random_seed=random_seed,
            )

            # Need to update random seed
            box._random_seed = init_boxes.random_seed

        # Create appropriate previous_spin_temp
        if not isinstance(previous_spin_temp, TsBox):
            if prev_z > global_params.Z_HEAT_MAX or prev_z is None:
                previous_spin_temp = TsBox(
                    redshift=global_params.Z_HEAT_MAX,
                    user_params=init_boxes.user_params,
                    cosmo_params=init_boxes.cosmo_params,
                    astro_params=astro_params,
                    flag_options=flag_options,
                )
            else:
                previous_spin_temp = spin_temperature(
                    init_boxes=init_boxes,
                    astro_params=astro_params,
                    flag_options=flag_options,
                    redshift=prev_z,
                    regenerate=regenerate,
                    write=write,
                    direc=direc,
                    cleanup=False,  # we know we'll need the memory again
                )

        # Dynamically produce the perturbed field.
        if perturbed_field is None or not perturbed_field.filled:
            perturbed_field = perturb_field(
                redshift=redshift,
                init_boxes=init_boxes,
                regenerate=regenerate,
                write=write,
                direc=direc,
            )

        # Run the C Code
        return _call_c_func(
            lib.ComputeTsBox,
            box,
            direc,
            redshift,
            previous_spin_temp.redshift,
            box.user_params,
            box.cosmo_params,
            box.astro_params,
            box.flag_options,
            perturbed_field.redshift,
            cleanup,
            perturbed_field,
            previous_spin_temp,
            write=write,
        )


def brightness_temperature(
    *,
    ionized_box,
    perturbed_field,
    spin_temp=None,
    write=None,
    regenerate=None,
    direc=None,
    **global_kwargs,
):
    r"""
    Compute a coeval brightness temperature box.

    Parameters
    ----------
    ionized_box: :class:`IonizedBox`
        A pre-computed ionized box.
    perturbed_field: :class:`PerturbedField`
        A pre-computed perturbed field at the same redshift as `ionized_box`.
    spin_temp: :class:`TsBox`, optional
        A pre-computed spin temperature, at the same redshift as the other boxes.
    \*\*global_kwargs :
        Any attributes for :class:`~py21cmfast.inputs.GlobalParams`. This will
        *temporarily* set global attributes for the duration of the function. Note that
        arguments will be treated as case-insensitive.

    Returns
    -------
    :class:`BrightnessTemp` instance.
    """
    direc, regenerate, write = _get_config_options(direc, regenerate, write)

    with global_params.use(**global_kwargs):
        _verify_types(
            perturbed_field=perturbed_field,
            spin_temp=spin_temp,
            ionized_box=ionized_box,
        )

        # don't ignore redshift here
        _check_compatible_inputs(ionized_box, perturbed_field, spin_temp, ignore=[])

        # ensure ionized_box and perturbed_field aren't None, as we don't do
        # any dynamic calculations here.
        if ionized_box is None or perturbed_field is None:
            raise ValueError("both ionized_box and perturbed_field must be specified.")

        if spin_temp is None:
            if ionized_box.flag_options.USE_TS_FLUCT:
                raise ValueError(
                    "You have USE_TS_FLUCT=True, but have not provided a spin_temp!"
                )

            # Make an unused dummy box.
            spin_temp = TsBox(redshift=0, dummy=True)

        box = BrightnessTemp(
            user_params=ionized_box.user_params,
            cosmo_params=ionized_box.cosmo_params,
            astro_params=ionized_box.astro_params,
            flag_options=ionized_box.flag_options,
            redshift=ionized_box.redshift,
            random_seed=ionized_box.random_seed,
        )

        # Check whether the boxes already exist on disk.
        if not regenerate:
            try:
                box.read(direc)
                logger.info(
                    "Existing brightness_temp box found and read in (seed=%s)."
                    % (box.random_seed)
                )
                return box
            except IOError:
                pass

        return _call_c_func(
            lib.ComputeBrightnessTemp,
            box,
            direc,
            ionized_box.redshift,
            ionized_box.user_params,
            ionized_box.cosmo_params,
            ionized_box.astro_params,
            ionized_box.flag_options,
            spin_temp,
            ionized_box,
            perturbed_field,
            write=write,
        )


def _logscroll_redshifts(min_redshift, z_step_factor, zmax):
    redshifts = [min_redshift]  # mult by 1.001 is probably bad...
    while redshifts[-1] < zmax:
        redshifts.append((redshifts[-1] + 1.0) * z_step_factor - 1.0)
    return redshifts[::-1]


class Coeval:
    """A full coeval box with all associated data."""

    def __init__(
        self,
        redshift: float,
        init_box: InitialConditions,
        perturb: PerturbedField,
        ib: IonizedBox,
        bt: BrightnessTemp,
        photon_nonconservation_data=None,
    ):
        _check_compatible_inputs(init_box, perturb, ib, bt, ignore=[])

        self.redshift = redshift
        self.init_struct = init_box
        self.perturb_struct = perturb
        self.ionization_struct = ib
        self.brightness_temp_struct = bt
        self.photon_nonconservation_data = photon_nonconservation_data

        # Expose all the fields of the structs to the surface of the Coeval object
        for box in [init_box, perturb, ib, bt]:
            for field in box.fieldnames:
                setattr(self, field, getattr(box, field))

    @property
    def user_params(self):
        """User params shared by all datasets."""
        return self.brightness_temp_struct.user_params

    @property
    def cosmo_params(self):
        """Cosmo params shared by all datasets."""
        return self.brightness_temp_struct.cosmo_params

    @property
    def flag_options(self):
        """Flag Options shared by all datasets."""
        return self.brightness_temp_struct.flag_options

    @property
    def astro_params(self):
        """Astro params shared by all datasets."""
        return self.brightness_temp_struct.astro_params

    @property
    def random_seed(self):
        """Random seed shared by all datasets."""
        return self.brightness_temp_struct.random_seed


def run_coeval(
    *,
    redshift=None,
    user_params=None,
    cosmo_params=None,
    astro_params=None,
    flag_options=None,
    regenerate=None,
    write=None,
    direc=None,
    init_box=None,
    perturb=None,
    use_interp_perturb_field=False,
    random_seed=None,
    cleanup=True,
    **global_kwargs,
):
    r"""
    Evaluate a coeval ionized box at a given redshift, or multiple redshifts.

    This is generally the easiest and most efficient way to generate a set of coeval cubes at a
    given set of redshift. It self-consistently deals with situations in which the field needs to be
    evolved, and does this with the highest memory-efficiency, only returning the desired redshift.
    All other calculations are by default stored in the on-disk cache so they can be re-used at a
    later time.

    .. note:: User-supplied redshift are *not* used as previous redshift in any scrolling,
              so that pristine log-sampling can be maintained.

    Parameters
    ----------
    redshift: array_like
        A single redshift, or multiple redshift, at which to return results. The minimum of these
        will define the log-scrolling behaviour (if necessary).
    user_params : :class:`~inputs.UserParams`, optional
        Defines the overall options and parameters of the run.
    cosmo_params : :class:`~inputs.CosmoParams` , optional
        Defines the cosmological parameters used to compute initial conditions.
    astro_params : :class:`~inputs.AstroParams` , optional
        The astrophysical parameters defining the course of reionization.
    flag_options : :class:`~inputs.FlagOptions` , optional
        Some options passed to the reionization routine.
    init_box : :class:`~InitialConditions`, optional
        If given, the user and cosmo params will be set from this object, and it will not
        be re-calculated.
    perturb : list of :class:`~PerturbedField`, optional
        If given, must be compatible with init_box. It will merely negate the necessity
        of re-calculating the perturb fields.
    use_interp_perturb_field : bool, optional
        Whether to use a single perturb field, at the lowest redshift of the lightcone,
        to determine all spin temperature fields. If so, this field is interpolated in
        the underlying C-code to the correct redshift. This is less accurate (and no more
        efficient), but provides compatibility with older versions of 21cmFAST.
    cleanup : bool, optional
        A flag to specify whether the C routine cleans up its memory before returning.
        Typically, if `spin_temperature` is called directly, you will want this to be
        true, as if the next box to be calculate has different shape, errors will occur
        if memory is not cleaned. Note that internally, this is set to False until the
        last iteration.
    \*\*global_kwargs :
        Any attributes for :class:`~py21cmfast.inputs.GlobalParams`. This will
        *temporarily* set global attributes for the duration of the function. Note that
        arguments will be treated as case-insensitive.

    Returns
    -------
    coevals : :class:`~Coeval`
        The full data for the Coeval class, with init boxes, perturbed fields, ionized boxes,
        brightness temperature, and potential data from the conservation of photons. If a
        single redshift was specified, it will return such a class. If multiple redshifts
        were passed, it will return a list of such classes.

    Other Parameters
    ----------------
    regenerate, write, direc, random_seed :
        See docs of :func:`initial_conditions` for more information.
    """
    with global_params.use(**global_kwargs):
        if redshift is None and perturb is None:
            raise ValueError("Either redshift or perturb must be given")

        if redshift is None and perturb is None:
            raise ValueError("Either redshift or perturb must be given")

        direc, regenerate, write = _get_config_options(direc, regenerate, write)

        # Ensure perturb is a list of boxes, not just one.
        if perturb is not None:
            if not hasattr(perturb, "__len__"):
                perturb = [perturb]
        else:
            perturb = []

        random_seed, user_params, cosmo_params = _configure_inputs(
            [
                ("random_seed", random_seed),
                ("user_params", user_params),
                ("cosmo_params", cosmo_params),
            ],
            init_box,
            *perturb,
        )

        user_params = UserParams(user_params)
        cosmo_params = CosmoParams(cosmo_params)
        flag_options = FlagOptions(flag_options)
        astro_params = AstroParams(astro_params, INHOMO_RECO=flag_options.INHOMO_RECO)

        if init_box is None:  # no need to get cosmo, user params out of it.
            init_box = initial_conditions(
                user_params=user_params,
                cosmo_params=cosmo_params,
                random_seed=random_seed,
                write=write,
                regenerate=regenerate,
                direc=direc,
            )

        if perturb:
            if redshift is not None:
                if not all(p.redshift == z for p, z in zip(perturb, redshift)):
                    raise ValueError(
                        "Input redshifts do not match perturb field redshifts"
                    )

            else:
                redshift = [p.redshift for p in perturb]

        if flag_options.PHOTON_CONS:
            calibrate_photon_cons(
                user_params,
                cosmo_params,
                astro_params,
                flag_options,
                init_box,
                regenerate,
                write,
                direc,
            )

        singleton = False
        if not hasattr(redshift, "__len__"):
            singleton = True
            redshift = [redshift]

        if not perturb:
            for z in redshift:
                perturb += [
                    perturb_field(
                        redshift=z,
                        init_boxes=init_box,
                        regenerate=regenerate,
                        write=write,
                        direc=direc,
                    )
                ]

        # Get the list of redshift we need to scroll through.
        if flag_options.INHOMO_RECO or flag_options.USE_TS_FLUCT:
            redshifts = _logscroll_redshifts(
                min(redshift),
                global_params.ZPRIME_STEP_FACTOR,
                global_params.Z_HEAT_MAX,
            )
        else:
            redshifts = [min(redshift)]

        # Add in the redshift defined by the user, and sort in order
        # Turn into a set so that exact matching user-set redshift
        # don't double-up with scrolling ones.
        redshifts += redshift
        redshifts = sorted(set(redshifts), reverse=True)

        ib_tracker = [0] * len(redshift)
        bt = [0] * len(redshift)
        st, ib = None, None  # At first we don't have any "previous" st or ib.
        logger.debug("redshifts: %s", redshifts)

        minarg = np.argmin(redshift)

        # Iterate through redshift from top to bottom
        for z in redshifts:

            if flag_options.USE_TS_FLUCT:
                logger.debug("PID={} doing spin temp for z={}".format(os.getpid(), z))
                st2 = spin_temperature(
                    redshift=z,
                    previous_spin_temp=st,
                    perturbed_field=(
                        perturb[minarg]
                        if use_interp_perturb_field
                        else (perturb[redshift.index(z)] if z in redshift else None)
                    ),
                    # remember that perturb field is interpolated, so no need to provide exact one.
                    astro_params=astro_params,
                    flag_options=flag_options,
                    regenerate=regenerate,
                    init_boxes=init_box,
                    write=write,
                    direc=direc,
                    cleanup=(
                        cleanup and z == redshifts[-1]
                    ),  # cleanup if its the last time through
                )

                if z not in redshift:
                    st = st2

            logger.debug("PID={} doing ionize box for z={}".format(os.getpid(), z))
            ib2 = ionize_box(
                redshift=z,
                previous_ionize_box=ib,
                init_boxes=init_box,
                perturbed_field=perturb[redshift.index(z)] if z in redshift else None,
                # perturb field *not* interpolated here.
                astro_params=astro_params,
                flag_options=flag_options,
                spin_temp=st2 if flag_options.USE_TS_FLUCT else None,
                regenerate=regenerate,
                write=write,
                direc=direc,
                cleanup=(
                    cleanup and z == redshifts[-1]
                ),  # cleanup if its the last time through
            )

            if z not in redshift:
                ib = ib2
            else:
                logger.debug(
                    "PID={} doing brightness temp for z={}".format(os.getpid(), z)
                )
                ib_tracker[redshift.index(z)] = ib2
                bt[redshift.index(z)] = brightness_temperature(
                    ionized_box=ib2,
                    perturbed_field=perturb[redshift.index(z)],
                    spin_temp=st2 if flag_options.USE_TS_FLUCT else None,
                    write=write,
                    direc=direc,
                    regenerate=regenerate,
                )

        if flag_options.PHOTON_CONS:
            photon_nonconservation_data = _get_photon_nonconservation_data()
        else:
            photon_nonconservation_data = None
        coevals = [
            Coeval(z, init_box, p, ib, _bt, photon_nonconservation_data)
            for z, p, ib, _bt in zip(redshift, perturb, ib_tracker, bt)
        ]

        # If a single redshift was passed, then pass back singletons.
        if singleton:
            coevals = coevals[0]

        logger.debug("PID={} RETURNING FROM COEVAL".format(os.getpid()))

        return coevals


class LightCone:
    """A full Lightcone with all associate evolved data."""

    def __init__(
        self,
        redshift,
        user_params,
        cosmo_params,
        astro_params,
        flag_options,
        brightness_temp,
        node_redshifts=None,
        global_xHI=None,
        global_brightness_temp=None,
        photon_nonconservation_data=None,
    ):
        self.redshift = redshift
        self.user_params = user_params
        self.cosmo_params = cosmo_params
        self.astro_params = astro_params
        self.flag_options = flag_options
        self.brightness_temp = brightness_temp

        self.node_redshifts = node_redshifts
        self.global_xHI = global_xHI
        self.global_brightness_temp = global_brightness_temp

        self.photon_nonconservation_data = photon_nonconservation_data

    @property
    def cell_size(self):
        """Cell size [Mpc] of the lightcone voxels."""
        return self.user_params.BOX_LEN / self.user_params.HII_DIM

    @property
    def lightcone_dimensions(self):
        """Lightcone size over each dimension -- tuple of (x,y,z) in Mpc."""
        return (
            self.user_params.BOX_LEN,
            self.user_params.BOX_LEN,
            self.n_slices * self.cell_size,
        )

    @property
    def shape(self):
        """Shape of the lightcone as a 3-tuple."""
        return self.brightness_temp.shape

    @property
    def n_slices(self):
        """Number of redshift slices in the lightcone."""
        return self.shape[-1]

    @property
    def lightcone_coords(self):
        """Co-ordinates [Mpc] of each cell along the redshift axis."""
        return np.linspace(0, self.lightcone_dimensions[-1], self.n_slices)

    @property
    def lightcone_distances(self):
        """Comoving distance to each cell along the redshift axis, from z=0."""
        return (
            self.cosmo_params.cosmo.comoving_distance(self.redshift).value
            + self.lightcone_coords
        )

    @property
    def lightcone_redshifts(self):
        """Redshift of each cell along the redshift axis."""
        return np.array(
            [
                z_at_value(self.cosmo_params.cosmo.comoving_distance, d * units.Mpc)
                for d in self.lightcone_distances
            ]
        )


def run_lightcone(
    *,
    redshift=None,
    max_redshift=None,
    user_params=None,
    cosmo_params=None,
    astro_params=None,
    flag_options=None,
    regenerate=None,
    write=None,
    direc=None,
    init_box=None,
    perturb=None,
    random_seed=None,
    use_interp_perturb_field=False,
    cleanup=True,
    **global_kwargs,
):
    r"""
    Evaluate a full lightcone ending at a given redshift.

    This is generally the easiest and most efficient way to generate a lightcone, though it can
    be done manually by using the lower-level functions which are called by this function.

    Parameters
    ----------
    redshift : float
        The minimum redshift of the lightcone.
    max_redshift : float, optional
        The maximum redshift at which to keep lightcone information. By default, this is equal to
        `z_heat_max`. Note that this is not *exact*, but will be typically slightly exceeded.
    user_params : `~UserParams`, optional
        Defines the overall options and parameters of the run.
    astro_params : :class:`~AstroParams`, optional
        Defines the astrophysical parameters of the run.
    cosmo_params : :class:`~CosmoParams`, optional
        Defines the cosmological parameters used to compute initial conditions.
    flag_options : :class:`~FlagOptions`, optional
        Options concerning how the reionization process is run, eg. if spin temperature
        fluctuations are required.
    init_box : :class:`~InitialConditions`, optional
        If given, the user and cosmo params will be set from this object, and it will not be
        re-calculated.
    perturb : list of :class:`~PerturbedField`, optional
        If given, must be compatible with init_box. It will merely negate the necessity of
        re-calculating the
        perturb fields. It will also be used to set the redshift if given.
    use_interp_perturb_field : bool, optional
        Whether to use a single perturb field, at the lowest redshift of the lightcone,
        to determine all spin temperature fields. If so, this field is interpolated in the
        underlying C-code to the correct redshift. This is less accurate (and no more efficient),
        but provides compatibility with older versions of 21cmFAST.
    cleanup : bool, optional
        A flag to specify whether the C routine cleans up its memory before returning.
        Typically, if `spin_temperature` is called directly, you will want this to be
        true, as if the next box to be calculate has different shape, errors will occur
        if memory is not cleaned. Note that internally, this is set to False until the
        last iteration.
    \*\*global_kwargs :
        Any attributes for :class:`~py21cmfast.inputs.GlobalParams`. This will
        *temporarily* set global attributes for the duration of the function. Note that
        arguments will be treated as case-insensitive.

    Returns
    -------
    lightcone : :class:`~LightCone`
        The lightcone object.

    Other Parameters
    ----------------
    regenerate, write, direc, random_seed
        See docs of :func:`initial_conditions` for more information.
    """
    direc, regenerate, write = _get_config_options(direc, regenerate, write)

    with global_params.use(**global_kwargs):
        random_seed, user_params, cosmo_params = _configure_inputs(
            [
                ("random_seed", random_seed),
                ("user_params", user_params),
                ("cosmo_params", cosmo_params),
            ],
            init_box,
            perturb,
        )

        user_params = UserParams(user_params)
        cosmo_params = CosmoParams(cosmo_params)
        flag_options = FlagOptions(flag_options)
        astro_params = AstroParams(astro_params, INHOMO_RECO=flag_options.INHOMO_RECO)

        if init_box is None:  # no need to get cosmo, user params out of it.
            init_box = initial_conditions(
                user_params=user_params,
                cosmo_params=cosmo_params,
                write=write,
                regenerate=regenerate,
                direc=direc,
                random_seed=random_seed,
            )

        redshift = configure_redshift(redshift, perturb)

        if perturb is None:
            # The perturb field that we get here is at the *final* redshift, and can be used in TsBox.
            perturb = perturb_field(
                redshift=redshift,
                init_boxes=init_box,
                regenerate=regenerate,
                direc=direc,
                write=write,
            )

        max_redshift = (
            global_params.Z_HEAT_MAX
            if (
                flag_options.INHOMO_RECO
                or flag_options.USE_TS_FLUCT
                or max_redshift is None
            )
            else max_redshift
        )

        if flag_options.PHOTON_CONS:
            calibrate_photon_cons(
                user_params,
                cosmo_params,
                astro_params,
                flag_options,
                init_box,
                regenerate,
                write,
                direc,
            )

        # Get the redshift through which we scroll and evaluate the ionization field.
        scrollz = _logscroll_redshifts(
            redshift, global_params.ZPRIME_STEP_FACTOR, max_redshift
        )

        d_at_redshift, lc_distances, n_lightcone = _setup_lightcone(
            cosmo_params,
            max_redshift,
            redshift,
            scrollz,
            user_params,
            global_params.ZPRIME_STEP_FACTOR,
        )

        lc = np.zeros(
            (user_params.HII_DIM, user_params.HII_DIM, n_lightcone), dtype=np.float32
        )

        scroll_distances = (
            cosmo_params.cosmo.comoving_distance(scrollz).value - d_at_redshift
        )

        # Iterate through redshift from top to bottom
        st, ib, bt = None, None, None
        lc_index = 0
        box_index = 0
        neutral_fraction = np.zeros(len(scrollz))
        global_signal = np.zeros(len(scrollz))

        for iz, z in enumerate(scrollz):
            # Best to get a perturb for this redshift, to pass to brightness_temperature
            this_perturb = perturb_field(
                redshift=z,
                init_boxes=init_box,
                regenerate=regenerate,
                direc=direc,
                write=write,
            )

            if flag_options.USE_TS_FLUCT:
                st2 = spin_temperature(
                    redshift=z,
                    previous_spin_temp=st,
                    astro_params=astro_params,
                    flag_options=flag_options,
                    perturbed_field=perturb
                    if use_interp_perturb_field
                    else this_perturb,
                    regenerate=regenerate,
                    init_boxes=init_box,
                    write=write,
                    direc=direc,
                    cleanup=(cleanup and iz == (len(scrollz) - 1)),
                )

            ib2 = ionize_box(
                redshift=z,
                previous_ionize_box=ib,
                init_boxes=init_box,
                perturbed_field=this_perturb,
                astro_params=astro_params,
                flag_options=flag_options,
                spin_temp=st2 if flag_options.USE_TS_FLUCT else None,
                regenerate=regenerate,
                write=write,
                direc=direc,
                cleanup=(cleanup and iz == (len(scrollz) - 1)),
            )

            bt2 = brightness_temperature(
                ionized_box=ib2,
                perturbed_field=this_perturb,
                spin_temp=st2 if flag_options.USE_TS_FLUCT else None,
                write=write,
                direc=direc,
                regenerate=regenerate,
            )

            # Save mean/global quantities
            neutral_fraction[iz] = np.mean(ib2.xH_box)
            global_signal[iz] = np.mean(bt2.brightness_temp)

            # Interpolate the lightcone
            if z < max_redshift:
                # Do linear interpolation only.
                prev_d = scroll_distances[iz - 1]
                this_d = scroll_distances[iz]

                # Get the cells that need to be filled on this iteration.
                these_distances = lc_distances[
                    np.logical_and(lc_distances < prev_d, lc_distances >= this_d)
                ]

                n = len(these_distances)
                ind = np.arange(-(box_index + n), -box_index)

                lc[:, :, -(lc_index + n) : n_lightcone - lc_index] = (
                    np.abs(this_d - these_distances)
                    * bt.brightness_temp.take(ind + n_lightcone, axis=2, mode="wrap")
                    + np.abs(prev_d - these_distances)
                    * bt2.brightness_temp.take(ind + n_lightcone, axis=2, mode="wrap")
                ) / (np.abs(prev_d - this_d))

                lc_index += n
                box_index += n

            # Save current ones as old ones.
            if flag_options.USE_TS_FLUCT:
                st = st2
            ib = ib2
            bt = bt2

        if flag_options.PHOTON_CONS:
            photon_nonconservation_data = _get_photon_nonconservation_data()
        else:
            photon_nonconservation_data = None

        return LightCone(
            redshift,
            user_params,
            cosmo_params,
            astro_params,
            flag_options,
            lc,
            node_redshifts=scrollz,
            global_xHI=neutral_fraction,
            global_brightness_temp=global_signal,
            photon_nonconservation_data=photon_nonconservation_data,
        )


def _setup_lightcone(
    cosmo_params, max_redshift, redshift, scrollz, user_params, z_step_factor
):
    # Here set up the lightcone box.
    # Get a length of the lightcone (bigger than it needs to be at first).
    d_at_redshift = cosmo_params.cosmo.comoving_distance(redshift).value
    Ltotal = (
        cosmo_params.cosmo.comoving_distance(scrollz[0] * z_step_factor).value
        - d_at_redshift
    )
    lc_distances = np.arange(0, Ltotal, user_params.BOX_LEN / user_params.HII_DIM)

    # Use max_redshift to get the actual distances we require.
    Lmax = cosmo_params.cosmo.comoving_distance(max_redshift).value - d_at_redshift
    first_greater = np.argwhere(lc_distances > Lmax)[0][0]

    # Get *at least* as far as max_redshift
    lc_distances = lc_distances[: (first_greater + 1)]

    n_lightcone = len(lc_distances)
    return d_at_redshift, lc_distances, n_lightcone


def _get_lightcone_redshifts(
    cosmo_params, max_redshift, redshift, user_params, z_step_factor
):
    scrollz = _logscroll_redshifts(redshift, z_step_factor, max_redshift)
    lc_distances = _setup_lightcone(
        cosmo_params, max_redshift, redshift, scrollz, user_params, z_step_factor
    )[1]
    lc_distances += cosmo_params.cosmo.comoving_distance(redshift).value

    return np.array(
        [
            z_at_value(cosmo_params.cosmo.comoving_distance, d * units.Mpc)
            for d in lc_distances
        ]
    )


def calibrate_photon_cons(
    user_params,
    cosmo_params,
    astro_params,
    flag_options,
    init_box,
    regenerate,
    write,
    direc,
    **global_kwargs,
):
    r"""
    Set up the photon non-conservation correction.

    Scrolls through in redshift, turning off all flag_options to construct a 21cmFAST calibration
    reionisation history to be matched to the analytic expression from solving the filling factor
    ODE.


    Parameters
    ----------
    user_params : `~UserParams`, optional
        Defines the overall options and parameters of the run.
    astro_params : :class:`~AstroParams`, optional
        Defines the astrophysical parameters of the run.
    cosmo_params : :class:`~CosmoParams`, optional
        Defines the cosmological parameters used to compute initial conditions.
    flag_options: :class:`~FlagOptions`, optional
        Options concerning how the reionization process is run, eg. if spin temperature
        fluctuations are required.
    z_step_factor: float, optional
        How large the logarithmic steps between redshift are (if required).
    init_box : :class:`~InitialConditions`, optional
        If given, the user and cosmo params will be set from this object, and it will not be
        re-calculated.
    \*\*global_kwargs :
        Any attributes for :class:`~py21cmfast.inputs.GlobalParams`. This will
        *temporarily* set global attributes for the duration of the function. Note that
        arguments will be treated as case-insensitive.

    Other Parameters
    ----------------
    regenerate, write
        See docs of :func:`initial_conditions` for more information.
    """
    direc, regenerate, write = _get_config_options(direc, regenerate, write)

    if not flag_options.PHOTON_CONS:
        return

    with global_params.use(**global_kwargs):
        # Create a new astro_params and flag_options just for the photon_cons correction
        astro_params_photoncons = deepcopy(astro_params)
        astro_params_photoncons._R_BUBBLE_MAX = astro_params.R_BUBBLE_MAX

        flag_options_photoncons = FlagOptions(
            USE_MASS_DEPENDENT_ZETA=flag_options.USE_MASS_DEPENDENT_ZETA,
            M_MIN_in_Mass=flag_options.M_MIN_in_Mass,
        )

        ib = None

        # Arrays for redshift and neutral fraction for the calibration curve
        z_for_photon_cons = []
        neutral_fraction_photon_cons = []

        # Initialise the analytic expression for the reionisation history
        _init_photon_conservation_correction(
            user_params=user_params,
            cosmo_params=cosmo_params,
            astro_params=astro_params,
            flag_options=flag_options,
        )

        # Determine the starting redshift to start scrolling through to create the
        # calibration reionisation history
        z = _calc_zstart_photon_cons()

        while z > 5.0:

            # Determine the ionisation box with recombinations, spin temperature etc.
            # turned off.
            this_perturb = perturb_field(
                redshift=z,
                init_boxes=init_box,
                regenerate=regenerate,
                write=write,
                direc=direc,
            )

            ib2 = ionize_box(
                redshift=z,
                previous_ionize_box=ib,
                init_boxes=init_box,
                perturbed_field=this_perturb,
                astro_params=astro_params_photoncons,
                flag_options=flag_options_photoncons,
                spin_temp=None,
                regenerate=regenerate,
                write=write,
                direc=direc,
            )

            mean_nf = np.mean(ib2.xH_box)

            # Save mean/global quantities
            neutral_fraction_photon_cons.append(mean_nf)
            z_for_photon_cons.append(z)

            # Can speed up sampling in regions where the evolution is slower
            if 0.01 < mean_nf <= 0.9:
                z -= 0.15
            else:
                z -= 0.5

            ib = ib2

        z_for_photon_cons = np.array(z_for_photon_cons[::-1])
        neutral_fraction_photon_cons = np.array(neutral_fraction_photon_cons[::-1])

        # Construct the spline for the calibration curve
        _calibrate_photon_conservation_correction(
            redshifts_estimate=z_for_photon_cons,
            nf_estimate=neutral_fraction_photon_cons,
            NSpline=len(z_for_photon_cons),
        )



import os

import pytest

import yaml
from click.testing import CliRunner

from py21cmfast import InitialConditions
from py21cmfast import UserParams
from py21cmfast import cli
from py21cmfast import query_cache


@pytest.fixture(scope="module")
def runner():
    return CliRunner()


@pytest.fixture(scope="module")
def user_params():
    return UserParams(HII_DIM=35, DIM=70, BOX_LEN=50)


@pytest.fixture(scope="module")
def cfg(user_params, tmpdirec):
    with open(os.path.join(tmpdirec.strpath, "cfg.yml"), "w") as f:
        yaml.dump({"user_params": user_params.self}, f)
    return os.path.join(tmpdirec.strpath, "cfg.yml")


def test_init(tmpdirec, user_params, runner, cfg):
    # Run the CLI
    result = runner.invoke(
        cli.main,
        ["init", "--direc", tmpdirec.strpath, "--seed", "101010", "--config", cfg],
    )

    if result.exception:
        print(result.output)

    assert result.exit_code == 0

    ic = InitialConditions(user_params=user_params, random_seed=101010)
    assert ic.exists(direc=tmpdirec.strpath)


def test_init_param_override(tmpdirec, runner, cfg):
    # Run the CLI
    result = runner.invoke(
        cli.main,
        [
            "init",
            "--direc",
            tmpdirec.strpath,
            "--seed",
            "102030",
            "--config",
            cfg,
            "--",
            "HII_DIM",
            "37",
            "DIM=52",
            "--OMm",
            "0.33",
        ],
    )
    assert result.exit_code == 0

    boxes = [
        res[1]
        for res in query_cache(
            direc=tmpdirec.strpath, kind="InitialConditions", seed=102030
        )
    ]

    assert len(boxes) == 1

    box = boxes[0]

    assert box.user_params.HII_DIM == 37
    assert box.user_params.DIM == 52
    assert box.cosmo_params.OMm == 0.33
    assert box.cosmo_params.cosmo.Om0 == 0.33


def test_perturb(tmpdirec, runner, cfg):
    # Run the CLI
    result = runner.invoke(
        cli.main,
        [
            "perturb",
            "35",
            "--direc",
            tmpdirec.strpath,
            "--seed",
            "101010",
            "--config",
            cfg,
        ],
    )
    assert result.exit_code == 0


def test_spin(tmpdirec, runner, cfg):
    # Run the CLI
    result = runner.invoke(
        cli.main,
        [
            "spin",
            "34.9",
            "--direc",
            tmpdirec.strpath,
            "--seed",
            "101010",
            "--config",
            cfg,
        ],
    )
    assert result.exit_code == 0


def test_spin_heatmax(tmpdirec, runner, cfg):
    # Run the CLI
    result = runner.invoke(
        cli.main,
        [
            "spin",
            "34.9",
            "--direc",
            tmpdirec.strpath,
            "--seed",
            "101010",
            "--config",
            cfg,
        ],
    )
    assert result.exit_code == 0


def test_ionize(tmpdirec, runner, cfg):
    # Run the CLI
    result = runner.invoke(
        cli.main,
        [
            "ionize",
            "35",
            "--direc",
            tmpdirec.strpath,
            "--seed",
            "101010",
            "--config",
            cfg,
        ],
    )
    assert result.exit_code == 0


def test_coeval(tmpdirec, runner, cfg):
    # Run the CLI
    result = runner.invoke(
        cli.main,
        [
            "coeval",
            "35",
            "--direc",
            tmpdirec.strpath,
            "--seed",
            "101010",
            "--config",
            cfg,
        ],
    )
    assert result.exit_code == 0


def test_lightcone(tmpdirec, runner, cfg):
    # Run the CLI
    result = runner.invoke(
        cli.main,
        [
            "lightcone",
            "30",
            "--direc",
            tmpdirec.strpath,
            "--seed",
            "101010",
            "--config",
            cfg,
            "-X",
            "35",
        ],
    )
    assert result.exit_code == 0


def test_query(tmpdirec, runner, cfg):
    # Run the CLI
    result = runner.invoke(
        cli.main, ["query", "--direc", tmpdirec.strpath, "--clear"]
    )  # Clear everything in tmpdirec
    assert result.exit_code == 0

    # Quickly run the default example once again.
    # Run the CLI
    result = runner.invoke(
        cli.main,
        ["init", "--direc", tmpdirec.strpath, "--seed", "101010", "--config", cfg],
    )
    assert result.exit_code == 0

    result = runner.invoke(
        cli.main, ["query", "--direc", tmpdirec.strpath, "--seed", "101010"]
    )

    assert result.output.startswith("1 Data Sets Found:")
    assert "random_seed:101010" in result.output
    assert "InitialConditions(" in result.output
